#!/usr/bin/env python
# encoding: utf-8
"""
rotor.py

Created by Andrew Ning on 2012-02-28.
Copyright (c)  NREL. All rights reserved.
"""

from __future__ import print_function
import numpy as np
import os
from openmdao.api import IndepVarComp, ExplicitComponent, Group, Problem
from scipy.optimize import minimize_scalar, minimize, brentq
from scipy.interpolate import PchipInterpolator

from wisdem.ccblade.ccblade_component import CCBladeGeometry, CCBladePower
from wisdem.ccblade import CCAirfoil, CCBlade

from wisdem.commonse.distribution import RayleighCDF, WeibullWithMeanCDF
from wisdem.commonse.utilities import vstack, trapz_deriv, linspace_with_deriv, smooth_min, smooth_abs
from wisdem.commonse.environment import PowerWind
from wisdem.commonse.akima import Akima
from wisdem.rotorse import RPM2RS, RS2RPM
from wisdem.rotorse.rotor_geometry import RotorGeometry
from wisdem.rotorse.rotor_geometry_yaml import ReferenceBlade
from wisdem.rotorse.rotor_fast import eval_unsteady

import time
# ---------------------
# Components
# ---------------------


class RegulatedPowerCurve(ExplicitComponent):

    def initialize(self):
        self.options.declare('naero')
        self.options.declare('n_pc')
        self.options.declare('n_pc_spline')
        self.options.declare('regulation_reg_II5',default=True)
        self.options.declare('regulation_reg_III',default=False)

        self.options.declare('n_aoa_grid')
        self.options.declare('n_Re_grid')

    
    def setup(self):
        naero       = self.naero = self.options['naero']
        n_pc        = self.options['n_pc']
        n_pc_spline = self.options['n_pc_spline']
        n_aoa_grid  = self.options['n_aoa_grid']
        n_Re_grid   = self.options['n_Re_grid']

        # parameters
        self.add_input('control_Vin',        val=0.0, units='m/s',  desc='cut-in wind speed')
        self.add_input('control_Vout',       val=0.0, units='m/s',  desc='cut-out wind speed')
        self.add_input('control_ratedPower', val=0.0, units='W',    desc='electrical rated power')
        self.add_input('control_minOmega',   val=0.0, units='rpm',  desc='minimum allowed rotor rotation speed')
        self.add_input('control_maxOmega',   val=0.0, units='rpm',  desc='maximum allowed rotor rotation speed')
        self.add_input('control_maxTS',      val=0.0, units='m/s',  desc='maximum allowed blade tip speed')
        self.add_input('control_tsr',        val=0.0,               desc='tip-speed ratio in Region 2 (should be optimized externally)')
        self.add_input('control_pitch',      val=0.0, units='deg',  desc='pitch angle in region 2 (and region 3 for fixed pitch machines)')
        self.add_discrete_input('drivetrainType',     val='GEARED')
        self.add_input('drivetrainEff',     val=0.0,               desc='overwrite drivetrain model with a given efficiency, used for FAST analysis')
        
        self.add_input('r',         val=np.zeros(naero), units='m',   desc='radial locations where blade is defined (should be increasing and not go all the way to hub or tip)')
        self.add_input('chord',     val=np.zeros(naero), units='m',   desc='chord length at each section')
        self.add_input('theta',     val=np.zeros(naero), units='deg', desc='twist angle at each section (positive decreases angle of attack)')
        self.add_input('Rhub',      val=0.0,             units='m',   desc='hub radius')
        self.add_input('Rtip',      val=0.0,             units='m',   desc='tip radius')
        self.add_input('hub_height',     val=0.0,             units='m',   desc='hub height')
        self.add_input('precone',   val=0.0,             units='deg', desc='precone angle', )
        self.add_input('tilt',      val=0.0,             units='deg', desc='shaft tilt', )
        self.add_input('yaw',       val=0.0,             units='deg', desc='yaw error', )
        self.add_input('precurve',      val=np.zeros(naero),    units='m', desc='precurve at each section')
        self.add_input('precurveTip',   val=0.0,                units='m', desc='precurve at tip')
        self.add_input('presweep',      val=np.zeros(naero),    units='m', desc='presweep at each section')
        self.add_input('presweepTip',   val=0.0,                units='m', desc='presweep at tip')
        
        # self.add_discrete_input('airfoils',  val=[0]*naero,                      desc='CCAirfoil instances')
        self.add_input('airfoils_cl', val=np.zeros((n_aoa_grid, naero, n_Re_grid)), desc='lift coefficients, spanwise')
        self.add_input('airfoils_cd', val=np.zeros((n_aoa_grid, naero, n_Re_grid)), desc='drag coefficients, spanwise')
        self.add_input('airfoils_cm', val=np.zeros((n_aoa_grid, naero, n_Re_grid)), desc='moment coefficients, spanwise')
        self.add_input('airfoils_aoa', val=np.zeros((n_aoa_grid)), units='deg', desc='angle of attack grid for polars')
        self.add_input('airfoils_Re', val=np.zeros((n_Re_grid)), desc='Reynolds numbers of polars')
        self.add_discrete_input('nBlades',         val=0,                              desc='number of blades')
        self.add_input('rho',       val=0.0,        units='kg/m**3',    desc='density of air')
        self.add_input('mu',        val=0.0,        units='kg/(m*s)',   desc='dynamic viscosity of air')
        self.add_input('shearExp',  val=0.0,                            desc='shear exponent')
        self.add_discrete_input('nSector',   val=4,                         desc='number of sectors to divide rotor face into in computing thrust and power')
        self.add_discrete_input('tiploss',   val=True,                      desc='include Prandtl tip loss model')
        self.add_discrete_input('hubloss',   val=True,                      desc='include Prandtl hub loss model')
        self.add_discrete_input('wakerotation', val=True,                   desc='include effect of wake rotation (i.e., tangential induction factor is nonzero)')
        self.add_discrete_input('usecd',     val=True,                      desc='use drag coefficient in computing induction factors')

        # outputs
        self.add_output('V',        val=np.zeros(n_pc), units='m/s',        desc='wind vector')
        self.add_output('Omega',    val=np.zeros(n_pc), units='rpm',        desc='rotor rotational speed')
        self.add_output('pitch',    val=np.zeros(n_pc), units='deg',        desc='rotor pitch schedule')
        self.add_output('P',        val=np.zeros(n_pc), units='W',          desc='rotor electrical power')
        self.add_output('T',        val=np.zeros(n_pc), units='N',          desc='rotor aerodynamic thrust')
        self.add_output('Q',        val=np.zeros(n_pc), units='N*m',        desc='rotor aerodynamic torque')
        self.add_output('M',        val=np.zeros(n_pc), units='N*m',        desc='blade root moment')
        self.add_output('Cp',       val=np.zeros(n_pc),                     desc='rotor electrical power coefficient')
        self.add_output('Cp_aero',  val=np.zeros(n_pc),                     desc='rotor aerodynamic power coefficient')
        self.add_output('Ct_aero',  val=np.zeros(n_pc),                     desc='rotor aerodynamic thrust coefficient')
        self.add_output('Cq_aero',  val=np.zeros(n_pc),                     desc='rotor aerodynamic torque coefficient')
        self.add_output('Cm_aero',  val=np.zeros(n_pc),                     desc='rotor aerodynamic moment coefficient')
        self.add_output('V_spline', val=np.zeros(n_pc_spline), units='m/s', desc='wind vector')
        self.add_output('P_spline', val=np.zeros(n_pc_spline), units='W',   desc='rotor electrical power')
        self.add_output('V_R25',       val=0.0,                units='m/s', desc='region 2.5 transition wind speed')
        self.add_output('rated_V',     val=0.0,                units='m/s', desc='rated wind speed')
        self.add_output('rated_Omega', val=0.0,                units='rpm', desc='rotor rotation speed at rated')
        self.add_output('rated_pitch', val=0.0,                units='deg', desc='pitch setting at rated')
        self.add_output('rated_T',     val=0.0,                units='N',   desc='rotor aerodynamic thrust at rated')
        self.add_output('rated_Q',     val=0.0,                units='N*m', desc='rotor aerodynamic torque at rated')
        self.add_output('ax_induct_cutin',   val=np.zeros(naero),           desc='rotor axial induction at cut-in wind speed along blade span')
        self.add_output('tang_induct_cutin', val=np.zeros(naero),           desc='rotor tangential induction at cut-in wind speed along blade span')
        self.add_output('aoa_cutin',val=np.zeros(naero),       units='deg', desc='angle of attack distribution along blade span at cut-in wind speed')
        self.add_output('cl_cutin', val=np.zeros(naero),                    desc='lift coefficient distribution along blade span at cut-in wind speed')
        self.add_output('cd_cutin', val=np.zeros(naero),                    desc='drag coefficient distribution along blade span at cut-in wind speed')

        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)
        
    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        # Create Airfoil class instances
        af = [None]*self.naero
        for i in range(self.naero):
            af[i] = CCAirfoil(inputs['airfoils_aoa'], inputs['airfoils_Re'], inputs['airfoils_cl'][:,i,:], inputs['airfoils_cd'][:,i,:], inputs['airfoils_cm'][:,i,:])
        

        self.ccblade = CCBlade(inputs['r'], inputs['chord'], inputs['theta'], af, inputs['Rhub'], inputs['Rtip'], discrete_inputs['nBlades'], inputs['rho'], inputs['mu'], inputs['precone'], inputs['tilt'], inputs['yaw'], inputs['shearExp'], inputs['hub_height'], discrete_inputs['nSector'], inputs['precurve'], inputs['precurveTip'],inputs['presweep'], inputs['presweepTip'], discrete_inputs['tiploss'], discrete_inputs['hubloss'],discrete_inputs['wakerotation'], discrete_inputs['usecd'])

        npc      = self.options['n_pc']
        Uhub     = np.linspace(inputs['control_Vin'],inputs['control_Vout'], npc).flatten()
        
        P_aero   = np.zeros( Uhub.shape )
        Cp_aero  = np.zeros( Uhub.shape )
        Ct_aero  = np.zeros( Uhub.shape )
        Cq_aero  = np.zeros( Uhub.shape )
        Cm_aero  = np.zeros( Uhub.shape )
        P        = np.zeros( Uhub.shape )
        Cp       = np.zeros( Uhub.shape )
        T        = np.zeros( Uhub.shape )
        Q        = np.zeros( Uhub.shape )
        M        = np.zeros( Uhub.shape )
        pitch    = np.zeros( Uhub.shape ) + inputs['control_pitch']

        # Unpack variables
        P_rated   = inputs['control_ratedPower']
        R_tip     = inputs['Rtip']
        tsr       = inputs['control_tsr']
        driveType = discrete_inputs['drivetrainType']
        driveEta  = inputs['drivetrainEff']
        
        # Set rotor speed based on TSR
        Omega_tsr = Uhub * tsr / R_tip

        # Determine maximum rotor speed (rad/s)- either by TS or by control input        
        Omega_max = min([inputs['control_maxTS'] / R_tip, inputs['control_maxOmega']*np.pi/30.])

        # Apply maximum and minimum rotor speed limits
        Omega     = np.maximum(np.minimum(Omega_tsr, Omega_max), inputs['control_minOmega']*np.pi/30.)
        Omega_rpm = Omega * 30. / np.pi

        # Set baseline power production
        P_aero, T, Q, M, Cp_aero, Ct_aero, Cq_aero, Cm_aero = self.ccblade.evaluate(Uhub, Omega_rpm, pitch, coefficients=True)
        P, eff  = CSMDrivetrain(P_aero, P_rated, driveType, driveEta)
        Cp      = Cp_aero*eff

        
        # Find Region 3 index
        region_bool = np.nonzero(P >= P_rated)[0]
        if len(region_bool)==0:
            i_3     = npc
            region3 = False
        else:
            i_3     = region_bool[0] + 1
            region3 = True

        # Guess at Region 2.5, but we will do a more rigorous search below
        if Omega_max < Omega_tsr[-1]:
            U_2p5 = np.interp(Omega[-1], Omega_tsr, Uhub)
            outputs['V_R25'] = U_2p5
        else:
            U_2p5 = Uhub[-1]
        i_2p5   = np.nonzero(U_2p5 <= Uhub)[0][0]

        # Find rated index and guess at rated speed
        if P_aero[-1] > P_rated:
            U_rated = np.interp(P_rated, P_aero*eff, Uhub)
        else:
            U_rated = Uhub[-1]
        i_rated = np.nonzero(U_rated <= Uhub)[0][0]
        
        # Function to be used inside of power maximization until Region 3
        def maximizePower(pitch, Uhub, Omega_rpm):
            P, _, _, _ = self.ccblade.evaluate([Uhub], [Omega_rpm], [pitch], coefficients=False)
            return -P

        # Maximize power until Region 3
        region2p5 = False
        for i in range(i_3):
            # No need to optimize if already doing well
            if Omega[i] == Omega_tsr[i]: continue

            # Find pitch value that gives highest power rating
            pitch0   = pitch[i] if i==0 else pitch[i-1]
            bnds     = [pitch0-10., pitch0+10.]
            pitch[i] = minimize_scalar(lambda x: maximizePower(x, Uhub[i], Omega_rpm[i]),
                                       bounds=bnds, method='bounded', options={'disp':False, 'xatol':1e-2, 'maxiter':40})['x']

            # Find associated power
            P_aero[i], T[i], Q[i], M[i], Cp_aero[i], Ct_aero[i], Cq_aero[i], Cm_aero[i] = self.ccblade.evaluate([Uhub[i]], [Omega_rpm[i]], [pitch[i]], coefficients=True)
            P[i], eff  = CSMDrivetrain(P_aero[i], P_rated, driveType, driveEta)
            Cp[i]      = Cp_aero[i]*eff

            # Note if we find Region 2.5
            if ( (not region2p5) and (Omega[i] == Omega_max) and (P[i] < P_rated) ):
                region2p5 = True
                i_2p5     = i

            # Stop if we find Region 3 early
            if P[i] > P_rated:
                i_3     = i+1
                i_rated = i
                break

            
        # Solve for rated velocity
        i = i_rated
        if i < npc-1:
            def const_Urated(x):
                pitch   = x[0]           
                Uhub_i  = x[1]
                Omega_i = min([Uhub_i * tsr / R_tip, Omega_max])
                P_aero_i, _, _, _ = self.ccblade.evaluate([Uhub_i], [Omega_i*30./np.pi], [pitch], coefficients=False)
                P_i,eff           = CSMDrivetrain(P_aero_i.flatten(), P_rated, driveType, driveEta)
                return (P_i - P_rated)

            if region2p5 and self.options['regulation_reg_II5']:
                # Have to search over both pitch and speed
                x0            = [0.0, Uhub[i]]
                bnds          = [ np.sort([pitch[i-1], pitch[i+1]]), [Uhub[i-1], Uhub[i+1]] ]
                const         = {}
                const['type'] = 'eq'
                const['fun']  = const_Urated
                params_rated  = minimize(lambda x: x[1], x0, method='slsqp', bounds=bnds, constraints=const, tol=1e-3)

                if params_rated.success and not np.isnan(params_rated.x[1]):
                    U_rated  = params_rated.x[1]
                    pitch[i] = params_rated.x[0]
                else:
                    U_rated = U_rated # Use guessed value earlier
                    pitch[i] = 0.0
            else:
                # Just search over speed
                pitch[i] = 0.0
                try:
                    U_rated = brentq(lambda x: const_Urated([0.0, x]), Uhub[i-1], Uhub[i+1],
                                     xtol = 1e-4, rtol = 1e-5, maxiter=40, disp=False)
                except ValueError:
                    U_rated = minimize_scalar(lambda x: np.abs(const_Urated([0.0, x])), bounds=[Uhub[i-1], Uhub[i+1]],
                                              method='bounded', options={'disp':False, 'xatol':1e-3, 'maxiter':40})['x']

            Omega_rated  = min([U_rated * tsr / R_tip, Omega_max])
            Omega[i:]    = np.minimum(Omega[i:], Omega_rated) # Stay at this speed if hit rated too early
            Omega_rpm    = Omega * 30. / np.pi
            P_aero[i], T[i], Q[i], M[i], Cp_aero[i], Ct_aero[i], Cq_aero[i], Cm_aero[i] = self.ccblade.evaluate([U_rated], [Omega_rpm[i]], [pitch[i]], coefficients=True)
            P[i], eff    = CSMDrivetrain(P_aero[i], P_rated, driveType, driveEta)
            Cp[i]        = Cp_aero[i]*eff
            
            # Store rated speed in array
            P[i_rated]    = P_rated
            Uhub[i_rated] = U_rated

        # Store outputs
        outputs['rated_V']     = np.float64(U_rated)
        outputs['rated_Omega'] = Omega_rpm[i]
        outputs['rated_pitch'] = pitch[i]
        outputs['rated_T']     = T[i]
        outputs['rated_Q']     = Q[i]

        
        if region3:
            # Function to be used to stay at rated power in Region 3
            def rated_power_dist(pitch, Uhub, Omega_rpm):
                P_aero, _, _, _ = self.ccblade.evaluate([Uhub], [Omega_rpm], [pitch], coefficients=False)
                P, eff          = CSMDrivetrain(P_aero, P_rated, driveType, driveEta)
                return (P - P_rated)

            # Solve for Region 3 pitch
            options = {'disp':False}
            if self.options['regulation_reg_III']:
                for i in range(i_3, npc):
                    pitch0   = pitch[i-1]
                    try:
                        pitch[i] = brentq(lambda x: rated_power_dist(x, Uhub[i], Omega_rpm[i]), pitch0, pitch0+10.,
                                          xtol = 1e-4, rtol = 1e-5, maxiter=40, disp=False)
                    except ValueError:
                        pitch[i] = minimize_scalar(lambda x: np.abs(rated_power_dist(x, Uhub[i], Omega_rpm[i])), bounds=[pitch0-5., pitch0+15.],
                                                  method='bounded', options={'disp':False, 'xatol':1e-3, 'maxiter':40})['x']

                    P_aero[i], T[i], Q[i], M[i], Cp_aero[i], Ct_aero[i], Cq_aero[i], Cm_aero[i] = self.ccblade.evaluate([Uhub[i]], [Omega_rpm[i]], [pitch[i]], coefficients=True)
                    P[i], eff  = CSMDrivetrain(P_aero[i], P_rated, driveType, driveEta)
                    Cp[i]      = Cp_aero[i]*eff
                    #P[i]       = P_rated

            else:
                P[i_3:]       = P_rated
                T[i_3:]       = 0
                Q[i_3:]       = P[i_3:] / Omega[i_3:]
                M[i_3:]       = 0
                pitch[i_3:]   = 0
                Cp[i_3:]      = P[i_3:] / (0.5 * inputs['rho'] * np.pi * R_tip**2 * Uhub[i_3:]**3)
                Ct_aero[i_3:] = 0
                Cq_aero[i_3:] = 0
                Cm_aero[i_3:] = 0

                    
        outputs['T']       = T
        outputs['Q']       = Q
        outputs['Omega']   = Omega_rpm

        outputs['P']       = P  
        outputs['Cp']      = Cp  
        outputs['Cp_aero'] = Cp_aero
        outputs['Ct_aero'] = Ct_aero
        outputs['Cq_aero'] = Cq_aero
        outputs['Cm_aero'] = Cm_aero
        outputs['V']       = Uhub
        outputs['M']       = M
        outputs['pitch']   = pitch
                
        self.ccblade.induction_inflow = True
        a_regII, ap_regII, alpha_regII, cl_regII, cd_regII = self.ccblade.distributedAeroLoads(Uhub[0], Omega_rpm[0], pitch[0], 0.0)
        
        # Fit spline to powercurve for higher grid density
        spline   = PchipInterpolator(Uhub, P)
        V_spline = np.linspace(inputs['control_Vin'], inputs['control_Vout'], self.options['n_pc_spline'])
        P_spline = spline(V_spline)
        
        # outputs
        outputs['V_spline']          = V_spline.flatten()
        outputs['P_spline']          = P_spline.flatten()
        outputs['ax_induct_cutin']   = a_regII
        outputs['tang_induct_cutin'] = ap_regII
        outputs['aoa_cutin']         = alpha_regII
        outputs['cl_cutin']          = cl_regII
        outputs['cd_cutin']          = cd_regII


class Cp_Ct_Cq_Tables(ExplicitComponent):
    def initialize(self):
        self.options.declare('naero')
        self.options.declare('n_pitch', default=20)
        self.options.declare('n_tsr', default=20)
        self.options.declare('n_U', default=1)
        self.options.declare('n_aoa_grid')
        self.options.declare('n_Re_grid')

    def setup(self):
        naero       = self.naero = self.options['naero']
        n_aoa_grid  = self.options['n_aoa_grid']
        n_Re_grid   = self.options['n_Re_grid']
        n_pitch     = self.options['n_pitch']
        n_tsr       = self.options['n_tsr']
        n_U         = self.options['n_U']
        
        # parameters        
        self.add_input('control_Vin',   val=0.0,             units='m/s',       desc='cut-in wind speed')
        self.add_input('control_Vout',  val=0.0,             units='m/s',       desc='cut-out wind speed')
        self.add_input('r',             val=np.zeros(naero), units='m',         desc='radial locations where blade is defined (should be increasing and not go all the way to hub or tip)')
        self.add_input('chord',         val=np.zeros(naero), units='m',         desc='chord length at each section')
        self.add_input('theta',         val=np.zeros(naero), units='deg',       desc='twist angle at each section (positive decreases angle of attack)')
        self.add_input('Rhub',          val=0.0,             units='m',         desc='hub radius')
        self.add_input('Rtip',          val=0.0,             units='m',         desc='tip radius')
        self.add_input('hub_height',    val=0.0,             units='m',         desc='hub height')
        self.add_input('precone',       val=0.0,             units='deg',       desc='precone angle')
        self.add_input('tilt',          val=0.0,             units='deg',       desc='shaft tilt')
        self.add_input('yaw',           val=0.0,             units='deg',       desc='yaw error')
        self.add_input('precurve',      val=np.zeros(naero), units='m',         desc='precurve at each section')
        self.add_input('precurveTip',   val=0.0,             units='m',         desc='precurve at tip')
        self.add_input('presweep',      val=np.zeros(naero), units='m',         desc='presweep at each section')
        self.add_input('presweepTip',   val=0.0,             units='m',         desc='presweep at tip')
        self.add_input('rho',           val=0.0,             units='kg/m**3',   desc='density of air')
        self.add_input('mu',            val=0.0,             units='kg/(m*s)',  desc='dynamic viscosity of air')
        self.add_input('shearExp',      val=0.0,                                desc='shear exponent')
        # self.add_discrete_input('airfoils',      val=[0]*naero,                 desc='CCAirfoil instances')
        self.add_input('airfoils_cl', val=np.zeros((n_aoa_grid, naero, n_Re_grid)), desc='lift coefficients, spanwise')
        self.add_input('airfoils_cd', val=np.zeros((n_aoa_grid, naero, n_Re_grid)), desc='drag coefficients, spanwise')
        self.add_input('airfoils_cm', val=np.zeros((n_aoa_grid, naero, n_Re_grid)), desc='moment coefficients, spanwise')
        self.add_input('airfoils_aoa', val=np.zeros((n_aoa_grid)), units='deg', desc='angle of attack grid for polars')
        self.add_input('airfoils_Re', val=np.zeros((n_Re_grid)), desc='Reynolds numbers of polars')
        self.add_discrete_input('nBlades',       val=0,                         desc='number of blades')
        self.add_discrete_input('nSector',       val=4,                         desc='number of sectors to divide rotor face into in computing thrust and power')
        self.add_discrete_input('tiploss',       val=True,                      desc='include Prandtl tip loss model')
        self.add_discrete_input('hubloss',       val=True,                      desc='include Prandtl hub loss model')
        self.add_discrete_input('wakerotation',  val=True,                      desc='include effect of wake rotation (i.e., tangential induction factor is nonzero)')
        self.add_discrete_input('usecd',         val=True,                      desc='use drag coefficient in computing induction factors')
        self.add_input('pitch_vector_in',  val=np.zeros(n_pitch), units='deg',  desc='pitch vector specified by the user')
        self.add_input('tsr_vector_in',    val=np.zeros(n_tsr),                 desc='tsr vector specified by the user')
        self.add_input('U_vector_in',      val=np.zeros(n_U),     units='m/s',  desc='wind vector specified by the user')

        # outputs
        self.add_output('Cp_aero_table',   val=np.zeros((n_tsr, n_pitch, n_U)), desc='table of aero power coefficient')
        self.add_output('Ct_aero_table',   val=np.zeros((n_tsr, n_pitch, n_U)), desc='table of aero thrust coefficient')
        self.add_output('Cq_aero_table',   val=np.zeros((n_tsr, n_pitch, n_U)), desc='table of aero torque coefficient')
        self.add_output('pitch_vector',    val=np.zeros(n_pitch), units='deg',  desc='pitch vector used')
        self.add_output('tsr_vector',      val=np.zeros(n_tsr),                 desc='tsr vector used')
        self.add_output('U_vector',        val=np.zeros(n_U),     units='m/s',  desc='wind vector used')
        
    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        # Create Airfoil class instances
        af = [None]*self.naero
        for i in range(self.naero):
            af[i] = CCAirfoil(inputs['airfoils_aoa'], inputs['airfoils_Re'], inputs['airfoils_cl'][:,i,:], inputs['airfoils_cd'][:,i,:], inputs['airfoils_cm'][:,i,:])
       

        n_pitch  = self.options['n_pitch']
        n_tsr    = self.options['n_tsr']
        n_U      = self.options['n_U']
        U_vector = inputs['U_vector_in']
        V_in     = inputs['control_Vin']
        V_out    = inputs['control_Vout']
        
        tsr_vector = inputs['tsr_vector_in']
        pitch_vector = inputs['pitch_vector_in']
        
        self.ccblade = CCBlade(inputs['r'], inputs['chord'], inputs['theta'], af, inputs['Rhub'], inputs['Rtip'], discrete_inputs['nBlades'], inputs['rho'], inputs['mu'], inputs['precone'], inputs['tilt'], inputs['yaw'], inputs['shearExp'], inputs['hub_height'], discrete_inputs['nSector'], inputs['precurve'], inputs['precurveTip'],inputs['presweep'], inputs['presweepTip'], discrete_inputs['tiploss'], discrete_inputs['hubloss'],discrete_inputs['wakerotation'], discrete_inputs['usecd'])
        
        if max(U_vector) == 0.:
            U_vector    = np.linspace(V_in[0],V_out[0], n_U)
        if max(tsr_vector) == 0.:
            tsr_vector = np.linspace(7.,11., n_tsr)
        if max(pitch_vector) == 0.:
            pitch_vector = np.linspace(-5., 5., n_pitch)
        
        outputs['pitch_vector'] = pitch_vector
        outputs['tsr_vector']   = tsr_vector        
        outputs['U_vector']     = U_vector
                
        R = inputs['Rtip']
        
        Cp_aero_table = np.zeros((n_tsr, n_pitch, n_U))
        Ct_aero_table = np.zeros((n_tsr, n_pitch, n_U))
        Cq_aero_table = np.zeros((n_tsr, n_pitch, n_U))
        
        for i in range(n_U):
            for j in range(n_tsr):
                U     =  U_vector[i] * np.ones(n_pitch)
                Omega = tsr_vector[j] *  U_vector[i] / R * 30. / np.pi * np.ones(n_pitch)
                _, _, _, _, outputs['Cp_aero_table'][j,:,i], outputs['Ct_aero_table'][j,:,i], outputs['Cq_aero_table'][j,:,i], _ = self.ccblade.evaluate(U, Omega, pitch_vector, coefficients=True)


# Class to define a constraint so that the blade cannot operate in stall conditions
class NoStallConstraint(ExplicitComponent):
    def initialize(self):
        
        self.options.declare('RefBlade')
        self.options.declare('verbosity', default = False)
    
    def setup(self):
        RefBlade    = self.options['RefBlade']
        NPTS        = len(RefBlade['pf']['s'])
        n_aoa_grid  = len(RefBlade['airfoils_aoa'])
        n_Re_grid   = len(RefBlade['airfoils_Re'])
        
        self.add_input('stall_angle_along_span', val=np.zeros(NPTS), units = 'deg', desc = 'Stall angle along blade span')
        self.add_input('aoa_along_span',         val=np.zeros(NPTS), units = 'deg', desc = 'Angle of attack along blade span')
        self.add_input('stall_margin',           val=0.0,            units = 'deg', desc = 'Minimum margin from the stall angle')
        self.add_input('min_s',                  val=0.0,            desc = 'Minimum nondimensional coordinate along blade span where to define the constraint (blade root typically stalls)')
        self.add_input('airfoils_cl',       val=np.zeros((n_aoa_grid, NPTS, n_Re_grid)), desc='lift coefficients, spanwise')
        self.add_input('airfoils_cd',       val=np.zeros((n_aoa_grid, NPTS, n_Re_grid)), desc='drag coefficients, spanwise')
        self.add_input('airfoils_cm',       val=np.zeros((n_aoa_grid, NPTS, n_Re_grid)), desc='moment coefficients, spanwise')
        self.add_input('airfoils_aoa',      val=np.zeros((n_aoa_grid)), units='deg', desc='angle of attack grid for polars')
        
        self.add_output('no_stall_constraint',   val=np.zeros(NPTS), desc = 'Constraint, ratio between angle of attack plus a margin and stall angle')

    def compute(self, inputs, outputs):
        
        verbosity = self.options['verbosity']
        RefBlade  = self.options['RefBlade']
        
        i_min = np.argmin(abs(inputs['min_s'] - RefBlade['pf']['s']))
        
        for i in range(len(RefBlade['pf']['s'])):
            unsteady = eval_unsteady(inputs['airfoils_aoa'], inputs['airfoils_cl'][:,i,0], inputs['airfoils_cd'][:,i,0], inputs['airfoils_cm'][:,i,0])
            inputs['stall_angle_along_span'][i] = unsteady['alpha1']
            if inputs['stall_angle_along_span'][i] == 0:
                inputs['stall_angle_along_span'][i] = 1e-6 # To avoid nan
        
        for i in range(i_min, len(RefBlade['pf']['s'])):
            outputs['no_stall_constraint'][i] = (inputs['aoa_along_span'][i] + inputs['stall_margin']) / inputs['stall_angle_along_span'][i]
        
            if verbosity == True:
                if outputs['no_stall_constraint'][i] > 1:
                    print('Blade is stalling at span location %.2f %%' % (RefBlade['pf']['s'][i]*100.))


class AEP(ExplicitComponent):
    def initialize(self):
        self.options.declare('n_pc_spline')
    
    def setup(self):
        n_pc_spline = self.options['n_pc_spline']
        """integrate to find annual energy production"""

        # inputs
        self.add_input('CDF_V', val=np.zeros(n_pc_spline), units='m/s', desc='cumulative distribution function evaluated at each wind speed')
        self.add_input('P', val=np.zeros(n_pc_spline), units='W', desc='power curve (power)')
        self.add_input('lossFactor', val=0.0, desc='multiplicative factor for availability and other losses (soiling, array, etc.)')

        # outputs
        self.add_output('AEP', val=0.0, units='kW*h', desc='annual energy production')

        #self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

    def compute(self, inputs, outputs):

        lossFactor = inputs['lossFactor']
        P = inputs['P']
        CDF_V = inputs['CDF_V']
        factor = lossFactor/1e3*365.0*24.0
        outputs['AEP'] = factor*np.trapz(P, CDF_V)  # in kWh
        '''
        dAEP_dP, dAEP_dCDF = trapz_deriv(P, CDF_V)
        dAEP_dP *= factor
        dAEP_dCDF *= factor

        dAEP_dlossFactor = np.array([outputs['AEP']/lossFactor])
        self.J = {}
        self.J['AEP', 'CDF_V'] = np.reshape(dAEP_dCDF, (1, len(dAEP_dCDF)))
        self.J['AEP', 'P'] = np.reshape(dAEP_dP, (1, len(dAEP_dP)))
        self.J['AEP', 'lossFactor'] = dAEP_dlossFactor

    def compute_partials(self, inputs, J):
        J = self.J
        '''


def CSMDrivetrain(aeroPower, ratedPower, drivetrainType, drivetrainEff):

    if drivetrainEff == 0.0:
        drivetrainType = drivetrainType.upper()
        if drivetrainType == 'GEARED':
            constant = 0.01289
            linear = 0.08510
            quadratic = 0.0

        elif drivetrainType == 'SINGLE_STAGE':
            constant = 0.01331
            linear = 0.03655
            quadratic = 0.06107

        elif drivetrainType == 'MULTI_DRIVE':
            constant = 0.01547
            linear = 0.04463
            quadratic = 0.05790

        elif drivetrainType == 'PM_DIRECT_DRIVE':
            constant = 0.01007
            linear = 0.02000
            quadratic = 0.06899
        elif drivetrainType == 'CONSTANT_EFF':
            constant = 0.00
            linear = 0.07
            quadratic = 0.0
        
        Pbar0 = aeroPower / ratedPower

        # handle negative power case (with absolute value)
        Pbar1, dPbar1_dPbar0 = smooth_abs(Pbar0, dx=0.01)

        # truncate idealized power curve for purposes of efficiency calculation
        Pbar, dPbar_dPbar1, _ = smooth_min(Pbar1, 1.0, pct_offset=0.01)

        # compute efficiency
        eff = 1.0 - (constant/Pbar + linear + quadratic*Pbar)
    else:
        eff = drivetrainEff
        
    return aeroPower * eff, eff


class OutputsAero(ExplicitComponent):
    def initialize(self):
        self.options.declare('npts_coarse_power_curve')
    
    def setup(self):
        npts_coarse_power_curve = self.options['npts_coarse_power_curve']

        # --- outputs ---
        self.add_input('rated_Omega_in', val=0.0, units='rpm', desc='rotor rotation speed at rated')
        self.add_input('rated_pitch_in', val=0.0, units='deg', desc='pitch setting at rated')
        self.add_input('rated_T_in', val=0.0, units='N', desc='rotor aerodynamic thrust at rated')
        self.add_input('rated_Q_in', val=0.0, units='N*m', desc='rotor aerodynamic torque at rated')

        self.add_input('V_extreme50', val=0.0, units='m/s', desc='survival wind speed')
        self.add_input('T_extreme_in', val=0.0, units='N', desc='thrust at survival wind condition')
        self.add_input('Q_extreme_in', val=0.0, units='N*m', desc='thrust at survival wind condition')

        # --- outputs ---
        self.add_output('V_extreme', val=0.0, units='m/s', desc='survival wind speed')
        self.add_output('T_extreme', val=0.0, units='N', desc='thrust at survival wind condition')
        self.add_output('Q_extreme', val=0.0, units='N*m', desc='thrust at survival wind condition')

        #self.declare_partials('V_extreme', 'V_extreme50')
        #self.declare_partials('T_extreme', 'T_extreme_in')
        #self.declare_partials('Q_extreme', 'Q_extreme_in')

    def compute(self, inputs, outputs):
        outputs['V_extreme'] = inputs['V_extreme50']
        outputs['T_extreme'] = inputs['T_extreme_in']
        outputs['Q_extreme'] = inputs['Q_extreme_in']
        '''
    def compute_partials(self, inputs, J):
        J['V_extreme', 'V_extreme50'] = 1
        J['T_extreme', 'T_extreme_in'] = 1
        J['Q_extreme', 'Q_extreme_in'] = 1
        '''

        

class RotorAeroPower(Group):
    def initialize(self):
        self.options.declare('RefBlade')
        self.options.declare('npts_coarse_power_curve', default=20)
        self.options.declare('npts_spline_power_curve', default=200)
        self.options.declare('regulation_reg_II5',      default=True)
        self.options.declare('regulation_reg_III',      default=True)
        self.options.declare('flag_Cp_Ct_Cq_Tables',    default=True)
        self.options.declare('topLevelFlag',            default=False)
        self.options.declare('user_update_routine',     default=None)
    
    def setup(self):
        RefBlade = self.options['RefBlade']
        npts_coarse_power_curve     = self.options['npts_coarse_power_curve']
        npts_spline_power_curve     = self.options['npts_spline_power_curve']
        regulation_reg_II5          = self.options['regulation_reg_II5']
        regulation_reg_III          = self.options['regulation_reg_III']
        flag_Cp_Ct_Cq_Tables        = self.options['flag_Cp_Ct_Cq_Tables']
        topLevelFlag                = self.options['topLevelFlag']
        user_update_routine         = self.options['user_update_routine']
        NPTS                        = len(RefBlade['pf']['s'])
        NAFgrid                     = len(RefBlade['airfoils_aoa'])
        NRe                         = len(RefBlade['airfoils_Re'])

        aeroIndeps = IndepVarComp()
        aeroIndeps.add_output('wind_reference_height',  val=0.0, units='m',     desc='reference hub height for IEC wind speed (used in CDF calculation)')
        aeroIndeps.add_output('control_Vin',            val=0.0, units='m/s',   desc='cut-in wind speed')
        aeroIndeps.add_output('control_Vout',           val=0.0, units='m/s',   desc='cut-out wind speed')
        aeroIndeps.add_output('machine_rating',         val=0.0, units='W',     desc='rated power')
        aeroIndeps.add_output('control_minOmega',       val=0.0, units='rpm',   desc='minimum allowed rotor rotation speed')
        aeroIndeps.add_output('control_maxOmega',       val=0.0, units='rpm',   desc='maximum allowed rotor rotation speed')
        aeroIndeps.add_output('control_maxTS',          val=0.0, units='m/s',   desc='maximum allowed blade tip speed')
        aeroIndeps.add_output('control_tsr',            val=0.0,                desc='tip-speed ratio in Region 2 (should be optimized externally)')
        aeroIndeps.add_output('control_pitch',          val=0.0, units='deg',   desc='pitch angle in region 2 (and region 3 for fixed pitch machines)')
        aeroIndeps.add_discrete_output('drivetrainType', val='GEARED')
        aeroIndeps.add_output('AEP_loss_factor',        val=1.0,                desc='availability and other losses (soiling, array, etc.)')
        aeroIndeps.add_output('shape_parameter',        val=0.0)
        aeroIndeps.add_output('drivetrainEff',          val=0.0,                desc='overwrite drivetrain model with a given efficiency, used for FAST analysis')
        self.add_subsystem('aeroIndeps', aeroIndeps, promotes=['*'])
        
        # --- Rotor Aero & Power ---
        if topLevelFlag:
            sharedIndeps = IndepVarComp()
            sharedIndeps.add_output('hub_height',   val=0.0, units='m')
            sharedIndeps.add_output('rho',          val=1.225, units='kg/m**3')
            sharedIndeps.add_output('mu',           val=1.81e-5, units='kg/(m*s)')
            sharedIndeps.add_output('shearExp',     val=0.2)
            sharedIndeps.add_discrete_output('tiploss', True)
            sharedIndeps.add_discrete_output('hubloss', True)
            sharedIndeps.add_discrete_output('wakerotation', True)
            sharedIndeps.add_discrete_output('usecd', True)
            sharedIndeps.add_discrete_output('nSector', val=4, desc='number of sectors to divide rotor face into in computing thrust and power')
            self.add_subsystem('sharedIndeps', sharedIndeps, promotes=['*'])
            
            self.add_subsystem('rotorGeom', RotorGeometry(RefBlade=RefBlade, topLevelFlag=topLevelFlag, user_update_routine=user_update_routine), promotes=['*'])

        # self.add_subsystem('tipspeed', MaxTipSpeed())
        self.add_subsystem('powercurve', RegulatedPowerCurve(naero=NPTS,
                                                             n_pc=npts_coarse_power_curve,
                                                             n_pc_spline=npts_spline_power_curve,
                                                             regulation_reg_II5=regulation_reg_II5,
                                                             regulation_reg_III=regulation_reg_III,
                                                             n_aoa_grid=NAFgrid,
                                                             n_Re_grid=NRe), promotes=['*'])

        if flag_Cp_Ct_Cq_Tables:
            self.add_subsystem('cpctcq_tables',   Cp_Ct_Cq_Tables(naero=NPTS,n_aoa_grid=NAFgrid,n_Re_grid=NRe), promotes=['*'])
        
        self.add_subsystem('nostallconstraint', NoStallConstraint(RefBlade = RefBlade, verbosity = False), promotes=['airfoils_cl','airfoils_cd','airfoils_cm','airfoils_aoa','no_stall_constraint'])
        self.add_subsystem('wind', PowerWind(nPoints=1), promotes=['shearExp'])
        self.add_subsystem('cdf', WeibullWithMeanCDF(nspline=npts_spline_power_curve))
        #self.add_subsystem('cdf', RayleighCDF(nspline=npts_spline_power_curve))
        self.add_subsystem('aep', AEP(n_pc_spline=npts_spline_power_curve), promotes=['AEP'])

        self.add_subsystem('outputs_aero', OutputsAero(npts_coarse_power_curve=npts_coarse_power_curve), promotes=['*'])

        self.connect('machine_rating',  'control_ratedPower')
        
        # connections to nostallconstraint
        self.connect('aoa_cutin','nostallconstraint.aoa_along_span')
        
        # connections to wind
        if topLevelFlag:
            self.connect('V_mean', 'wind.Uref')
            self.connect('wind_zvec', 'wind.z')
        self.connect('wind_reference_height', 'wind.zref')

        # connections to cdf
        self.connect('V_spline', 'cdf.x')
        self.connect('wind.U', 'cdf.xbar', src_indices=[0])
        self.connect('shape_parameter', 'cdf.k')

        # connections to aep
        self.connect('cdf.F', 'aep.CDF_V')
        self.connect('P_spline', 'aep.P')
        self.connect('AEP_loss_factor', 'aep.lossFactor')


def Init_RotorAeropower_wRefBlade(rotor, blade):
    # === blade grid ===
    rotor['hubFraction']      = blade['config']['hubD']/2./blade['pf']['r'][-1] #0.025  # (Float): hub location as fraction of radius
    rotor['bladeLength']      = blade['ctrl_pts']['bladeLength'] #61.5  # (Float, m): blade length (if not precurved or swept) otherwise length of blade before curvature
    rotor['precone']          = blade['config']['cone_angle'] #2.5  # (Float, deg): precone angle
    rotor['tilt']             = blade['config']['tilt_angle'] #5.0  # (Float, deg): shaft tilt
    rotor['yaw']              = 0.0  # (Float, deg): yaw error
    rotor['nBlades']          = blade['config']['number_of_blades'] #3  # (Int): number of blades
    # ------------------
    
    # === blade geometry ===
    rotor['r_max_chord']      = blade['ctrl_pts']['r_max_chord']  # 0.23577 #(Float): location of max chord on unit radius
    rotor['chord_in']         = np.array(blade['ctrl_pts']['chord_in']) # np.array([3.2612, 4.3254, 4.5709, 3.7355, 2.69923333, 1.4621])  # (Array, m): chord at control points. defined at hub, then at linearly spaced locations from r_max_chord to tip
    rotor['theta_in']         = np.array(blade['ctrl_pts']['theta_in']) # np.array([0.0, 13.2783, 12.30514836,  6.95106536,  2.72696309, -0.0878099]) # (Array, deg): twist at control points.  defined at linearly spaced locations from r[idx_cylinder] to tip
    rotor['precurve_in']      = np.array(blade['ctrl_pts']['precurve_in']) #np.array([0.0, 0.0, 0.0])  # (Array, m): precurve at control points.  defined at same locations at chord, starting at 2nd control point (root must be zero precurve)
    rotor['presweep_in']      = np.array(blade['ctrl_pts']['presweep_in']) #np.array([0.0, 0.0, 0.0])  # (Array, m): precurve at control points.  defined at same locations at chord, starting at 2nd control point (root must be zero precurve)
    rotor['sparT_in']         = np.array(blade['ctrl_pts']['sparT_in']) # np.array([0.0, 0.05, 0.047754, 0.045376, 0.031085, 0.0061398])  # (Array, m): spar cap thickness parameters
    rotor['teT_in']           = np.array(blade['ctrl_pts']['teT_in']) # np.array([0.0, 0.1, 0.09569, 0.06569, 0.02569, 0.00569])  # (Array, m): trailing-edge thickness parameters
    # if 'le_var' in blade['precomp']['le_var']:
    #     rotor['leT_in']       = np.array(blade['ctrl_pts']['leT_in']) ## (Array, m): leading-edge thickness parameters
    rotor['airfoil_position'] = np.array(blade['outer_shape_bem']['airfoil_position']['grid'])
    # ------------------
    
    # === atmosphere ===
    rotor['rho']              = 1.225  # (Float, kg/m**3): density of air
    rotor['mu']               = 1.81206e-5  # (Float, kg/m/s): dynamic viscosity of air
    rotor['hub_height']       = blade['config']['hub_height']  # (Float, m): hub height
    rotor['shearExp']         = 0.25  # (Float): shear exponent
    rotor['shape_parameter']  = 2.0
    rotor['turbine_class']    = blade['config']['turbine_class'].upper() #TURBINE_CLASS['I']  # (Enum): IEC turbine class
    rotor['wind_reference_height'] = blade['config']['hub_height']  # (Float, m): hub height
    # ----------------------
    
    # === control ===
    rotor['control_Vin']      = blade['config']['Vin'] #3.0  # (Float, m/s): cut-in wind speed
    rotor['control_Vout']     = blade['config']['Vout'] #25.0  # (Float, m/s): cut-out wind speed
    rotor['machine_rating']   = blade['config']['rating'] #5e6  # (Float, W): rated power
    rotor['control_minOmega'] = blade['config']['minOmega'] #0.0  # (Float, rpm): minimum allowed rotor rotation speed
    rotor['control_maxOmega'] = blade['config']['maxOmega'] #12.0  # (Float, rpm): maximum allowed rotor rotation speed
    rotor['control_maxTS']    = blade['config']['maxTS']
    rotor['control_tsr']      = blade['config']['tsr'] #7.55  # (Float): tip-speed ratio in Region 2 (should be optimized externally)
    rotor['control_pitch']    = blade['config']['pitch'] #0.0  # (Float, deg): pitch angle in region 2 (and region 3 for fixed pitch machines)
    # ----------------------
    
    # === no stall constraint ===
    rotor['nostallconstraint.min_s']        = 0.25  # The stall constraint is only computed from this value (nondimensional coordinate along blade span) to blade tip
    rotor['nostallconstraint.stall_margin'] = 3.0   # Values in deg of stall margin
    # ----------------------
    
    # === aero and structural analysis options ===
    rotor['nSector'] = 4  # (Int): number of sectors to divide rotor face into in computing thrust and power
    rotor['AEP_loss_factor'] = 1.0  # (Float): availability and other losses (soiling, array, etc.)
    rotor['drivetrainType']   = blade['config']['drivetrain'].upper() #DRIVETRAIN_TYPE['GEARED']  # (Enum)
    # ----------------------
    return rotor

if __name__ == '__main__':


    tt = time.time()

    # Turbine Ontology input
    fname_input  = "turbine_inputs/nrel5mw_mod_update.yaml"
    # fname_output = "turbine_inputs/nrel5mw_mod_out.yaml"
    fname_schema = "turbine_inputs/IEAontology_schema.yaml"
    
    # Initialize blade design
    refBlade = ReferenceBlade()
    refBlade.verbose = True
    refBlade.NINPUT  = 5
    refBlade.NPTS    = 50
    refBlade.spar_var = ['Spar_Cap_SS', 'Spar_Cap_PS']
    refBlade.te_var   = 'TE_reinforcement'
    # refBlade.le_var   = 'LE_reinforcement'
    refBlade.validate     = False
    refBlade.fname_schema = fname_schema
    
    blade = refBlade.initialize(fname_input)
    rotor = Problem()
    npts_coarse_power_curve = 20 # (Int): number of points to evaluate aero analysis at
    npts_spline_power_curve = 2000  # (Int): number of points to use in fitting spline to power curve
    regulation_reg_II5      = True # calculate Region 2.5 pitch schedule, False will not maximize power in region 2.5
    regulation_reg_III      = True # calculate Region 3 pitch schedule, False will return erroneous Thrust, Torque, and Moment for above rated
    flag_Cp_Ct_Cq_Tables    = True # Compute Cp-Ct-Cq-Beta-TSR tables
    
    rotor.model = RotorAeroPower(RefBlade=blade,
                                 npts_coarse_power_curve=npts_coarse_power_curve,
                                 npts_spline_power_curve=npts_spline_power_curve,
                                 regulation_reg_II5=regulation_reg_II5,
                                 regulation_reg_III=regulation_reg_III,
                                 topLevelFlag=True)
    
    #rotor.setup(check=False)
    rotor.setup()
    rotor = Init_RotorAeropower_wRefBlade(rotor, blade)

    # === run and outputs ===
    rotor.run_driver()
    #rotor.check_partials(compact_print=True, step=1e-6, form='central')#, includes='*ae*')

    print('Run time = ', time.time()-tt)
    print('AEP =', rotor['AEP'])
    print('diameter =', rotor['diameter'])
    print('ratedConditions.V =', rotor['rated_V'])
    print('ratedConditions.Omega =', rotor['rated_Omega'])
    print('ratedConditions.pitch =', rotor['rated_pitch'])
    print('ratedConditions.T =', rotor['rated_T'])
    print('ratedConditions.Q =', rotor['rated_Q'])
    
    import matplotlib.pyplot as plt
    plt.figure()
    plt.plot(rotor['V'], rotor['P']/1e6)
    plt.xlabel('wind speed (m/s)')
    plt.xlabel('power (W)')
    plt.show()
    
    if flag_Cp_Ct_Cq_Tables:
        n_pitch = len(rotor['pitch_vector'])
        n_tsr   = len(rotor['tsr_vector'])
        n_U     = len(rotor['U_vector'])
        for i in range(n_U):
            fig0, ax0 = plt.subplots()
            CS0 = ax0.contour(rotor['pitch_vector'], rotor['tsr_vector'], rotor['Cp_aero_table'][:, :, i], levels=[0.0, 0.3, 0.40, 0.42, 0.44, 0.45, 0.46, 0.47, 0.48, 0.49, 0.50 ])
            ax0.clabel(CS0, inline=1, fontsize=12)
            plt.title('Power Coefficient', fontsize=14, fontweight='bold')
            plt.xlabel('Pitch Angle [deg]', fontsize=14, fontweight='bold')
            plt.ylabel('TSR [-]', fontsize=14, fontweight='bold')
            plt.xticks(fontsize=12)
            plt.yticks(fontsize=12)
            plt.grid(color=[0.8,0.8,0.8], linestyle='--')
            plt.subplots_adjust(bottom = 0.15, left = 0.15)

            fig0, ax0 = plt.subplots()
            CS0 = ax0.contour(rotor['pitch_vector'], rotor['tsr_vector'], rotor['Ct_aero_table'][:, :, i])
            ax0.clabel(CS0, inline=1, fontsize=12)
            plt.title('Thrust Coefficient', fontsize=14, fontweight='bold')
            plt.xlabel('Pitch Angle [deg]', fontsize=14, fontweight='bold')
            plt.ylabel('TSR [-]', fontsize=14, fontweight='bold')
            plt.xticks(fontsize=12)
            plt.yticks(fontsize=12)
            plt.grid(color=[0.8,0.8,0.8], linestyle='--')
            plt.subplots_adjust(bottom = 0.15, left = 0.15)

            
            fig0, ax0 = plt.subplots()
            CS0 = ax0.contour(rotor['pitch_vector'], rotor['tsr_vector'], rotor['Cq_aero_table'][:, :, i])
            ax0.clabel(CS0, inline=1, fontsize=12)
            plt.title('Torque Coefficient', fontsize=14, fontweight='bold')
            plt.xlabel('Pitch Angle [deg]', fontsize=14, fontweight='bold')
            plt.ylabel('TSR [-]', fontsize=14, fontweight='bold')
            plt.xticks(fontsize=12)
            plt.yticks(fontsize=12)
            plt.grid(color=[0.8,0.8,0.8], linestyle='--')
            plt.subplots_adjust(bottom = 0.15, left = 0.15)
            
            plt.show()
   
    



from .csystem import DirectionVector
from .utilities import cosd, sind, tand
from .SegIntersect import SegIntersect, CalcDist
from .Material import Material
from .tube import Tube
from .WindWaveDrag import AeroHydroLoads, CylinderWindDrag, CylinderWaveDrag
from .akima import Akima
from .enum import Enum
from .constants import gravity, eps
NFREQ  = 5



from __future__ import print_function
from floatingse.floating import FloatingSE
from openmdao.api import Problem, ScipyOptimizer, DumpRecorder
import numpy as np
import cPickle as pickle        
from StringIO import StringIO

TOWSEC = 3
NSECTIONS = 4
NPTS = 100
Ten_strings = ['DTU', 'DTU10', 'DTU10MW', '10', '10MW', 'DTU-10', 'DTU-10MW']
Five_strings = ['NREL', 'NREL5', 'NREL5MW', '5', '5MW', 'NREL-5', 'NREL-5MW']


def vecOption(x, in1s):
    myones = in1s if type(in1s) == type(np.array([])) else np.ones((in1s,))
    return (x*myones) if type(x)==type(0.0) or len(x) == 1 else x


def set_axes_equal(ax):
    '''Make axes of 3D plot have equal scale so that spheres appear as spheres,
    cubes as cubes, etc..  This is one possible solution to Matplotlib's
    ax.set_aspect('equal') and ax.axis('equal') not working for 3D.

    Input
      ax: a matplotlib axis, e.g., as output from plt.gca().
    '''

    x_limits = ax.get_xlim3d()
    y_limits = ax.get_ylim3d()
    z_limits = ax.get_zlim3d()

    x_range = abs(x_limits[1] - x_limits[0])
    x_middle = np.mean(x_limits)
    y_range = abs(y_limits[1] - y_limits[0])
    y_middle = np.mean(y_limits)
    z_range = abs(z_limits[1] - z_limits[0])
    z_middle = np.mean(z_limits)

    # The plot bounding box is a sphere in the sense of the infinity
    # norm, hence I call half the max range the plot radius.
    plot_radius = 0.5*max([x_range, y_range, z_range])

    ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
    ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
    ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])

class FloatingInstance(object):
    def __init__(self):
        self.prob = Problem()
        self.params = {}
        self.design_variables = []
        self.constraints = []
        self.optimizer = None
        self.objective = None

        # Environmental parameters
        self.params['water_depth']               = 200.0
        #self.params['air_density']              = 1.198
        self.params['main.windLoads.rho']        = 1.198
        #self.params['air_viscosity']            = 1.81e-5
        self.params['main.windLoads.mu']         = 1.81e-5
        self.params['water_density']             = 1025.0
        #self.params['water_viscosity']          = 8.9e-4
        self.params['main.waveLoads.mu']         = 8.9e-4
        #self.params['wave_height']              = 10.8
        self.params['Hs']                      = 10.8
        #self.params['wave_period']              = 9.8
        self.params['T']                         = 9.8
        self.params['Uc']                        = 0.0
        #self.params['wind_reference_speed']     = 11.0
        self.params['Uref']                      = 11.0
        #self.params['wind_reference_height']    = 90.0
        self.params['zref']                      = 119.0
        #self.params['alpha']                    = 0.11
        self.params['shearExp']                  = 0.11
        #self.params['morison_mass_coefficient'] = 2.0
        self.params['cm']                        = 2.0
        self.params['z0']                        = 0.0
        self.params['yaw']                       = 0.0
        self.params['beta']                      = 0.0
        self.params['cd_usr']                    = np.inf
        self.params['z_offset']                  = 0.0
        
        # Encironmental constaints
        self.params['wave_period_range_low']                = 2.0
        self.params['wave_period_range_high']               = 20.0
        
        # Mooring parameters
        self.params['max_offset']                   = 0.1*self.params['water_depth'] # Assumption        
        self.params['operational_heel']             = 6.0
        self.params['max_survival_heel']                    = 15.0
        self.params['mooring_type']                         = 'chain'
        self.params['anchor_type']                          = 'suctionpile'
        self.params['mooring_cost_rate']                    = 1.1
        self.params['number_of_mooring_connections']        = 3
        self.params['mooring_lines_per_connection']         = 1

        # Steel properties
        self.params['material_density']                     = 7850.0
        self.params['E']                                    = 200e9
        self.params['G']                                    = 79.3e9
        self.params['nu']                                   = 0.26
        self.params['yield_stress']                         = 3.45e8
        self.params['loading']                              = 'hydrostatic'

        # Design constraints
        self.params['max_draft']                      = 150.0
        self.params['max_taper_ratio']                      = 0.4
        self.params['min_diameter_thickness_ratio']         = 120.0

        # Safety factors
        self.params['gamma_f']                              = 1.35
        self.params['gamma_b']                              = 1.1
        self.params['gamma_m']                              = 1.1
        self.params['gamma_n']                              = 1.0
        self.params['gamma_fatigue']                        = 1.755
        
        # Typically static- set defaults
        self.params['permanent_ballast_density']            = 4492.0
        self.params['bulkhead_mass_factor']                 = 1.0
        self.params['ring_mass_factor']                     = 1.0
        self.params['shell_mass_factor']                    = 1.0
        self.params['column_mass_factor']                   = 1.05
        self.params['outfitting_mass_fraction']             = 0.06
        self.params['ballast_cost_rate']                    = 0.1
        self.params['material_cost_rate']                   = 1.1
        self.params['labor_cost_rate']                      = 1.0
        self.params['painting_cost_rate']                   = 14.4
        self.params['outfitting_cost_rate']                 = 1.5 * self.params['material_cost_rate']
        self.params['cross_attachment_pontoons_int']        = 1
        self.params['lower_attachment_pontoons_int']        = 1
        self.params['upper_attachment_pontoons_int']        = 1
        self.params['lower_ring_pontoons_int']              = 1
        self.params['upper_ring_pontoons_int']              = 1
        self.params['outer_cross_pontoons_int']             = 1 #False

        # OC4 Tower
        self.params['hub_height']                           = 90.0
        self.params['tower_outer_diameter']                 = np.linspace(6.5, 3.87, TOWSEC+1)
        self.params['tower_section_height']                 = vecOption(77.6/TOWSEC, TOWSEC)
        self.params['tower_wall_thickness']                 = np.linspace(0.027, 0.019, TOWSEC)
        self.params['tower_buckling_length']                = 30.0
        self.params['tower_outfitting_factor']              = 1.07
        self.params['hub_cm']                               = np.array([-5.01910, 0.0, 1.96256])
        self.params['tip_position']                         = np.array([-0.9315, 0.0, 27.5])
        self.params['rna_mass']                             = 350e3 #285598.8
        self.params['rna_I']                                = np.array([1.14930678e+08, 2.20354030e+07, 1.87597425e+07, 0.0, 5.03710467e+05, 0.0])
        self.params['rna_cg']                               = np.array([-1.13197635, 0.0, 0.50875268])
        # Max thrust
        self.params['rna_force']                            = np.array([1284744.196, 0.0,  -112400.5527])
        self.params['rna_moment']                           = np.array([3963732.762, 896380.8464,  -346781.6819])
        # Max wind speed
        #self.params['rna_force']                           = np.array([188038.8045, 0,  -16451.2637])
        #self.params['rna_moment']                          = np.array([0.0, 131196.8431,  0.0])
        self.params['main_bulkhead_thickness']              = 0.05*np.array([1, 1, 0, 0, 1]) # Locations/thickness of internal bulkheads at section interfaces [m]
        self.params['offset_bulkhead_thickness']         = 0.05*np.array([1, 1, 0, 0, 1]) # Locations/thickness of internal bulkheads at section interfaces [m]
        self.params['Rhub']                              = 1.125
        self.params['downwind'] = False
        
        # Typically design (start at OC4 semi)
        self.params['radius_to_offset_column']           = 28.867513459481287
        self.params['number_of_offset_columns']          = 3
        self.params['main_freeboard']                       = 10.0
        self.params['offset_freeboard']                  = 12.0
        self.params['fairlead_location']                    = 0.1875 # Want 14 m
        self.params['fairlead_offset_from_shell']           = 40.868-28.867513459481287-6.0
        self.params['main_outer_diameter']                  = 6.5
        self.params['main_wall_thickness']                  = 0.03
        self.params['offset_wall_thickness']             = 0.06
        self.params['main_permanent_ballast_height']        = 1.0
        self.params['main_stiffener_web_height']            = 0.1
        self.params['main_stiffener_web_thickness']         = 0.04
        self.params['main_stiffener_flange_width']          = 0.1
        self.params['main_stiffener_flange_thickness']      = 0.02
        self.params['main_stiffener_spacing']               = 0.4
        self.params['offset_permanent_ballast_height']   = 0.1
        self.params['offset_stiffener_web_height']       = 0.1
        self.params['offset_stiffener_web_thickness']    = 0.04
        self.params['offset_stiffener_flange_width']     = 0.1
        self.params['offset_stiffener_flange_thickness'] = 0.02
        self.params['offset_stiffener_spacing']          = 0.4
        self.params['fairlead_support_outer_diameter']      = 2*1.6
        self.params['fairlead_support_wall_thickness']      = 0.0175
        self.params['pontoon_outer_diameter']               = 2*1.6
        self.params['pontoon_wall_thickness']               = 0.0175
        self.params['connection_ratio_max']                 = 0.25
        self.params['main_pontoon_attach_lower']            = 0.1
        self.params['main_pontoon_attach_upper']            = 1.0
        self.params['main_buoyancy_tank_diameter']      = 0.0
        self.params['main_buoyancy_tank_height']        = 0.0
        self.params['main_buoyancy_tank_location']      = 0.0
        self.params['offset_buoyancy_tank_diameter']      = 0.0
        self.params['offset_buoyancy_tank_height']        = 0.0
        self.params['offset_buoyancy_tank_location']      = 0.0
        
        self.set_length_main( 30.0 )
        self.set_length_aux( 32.0 )

        self.params['offset_outer_diameter']             = 2*6.0
        self.params['offset_buoyancy_tank_diameter']       = 24.0
        self.params['offset_buoyancy_tank_height']         = 5.0

        self.params['mooring_line_length']                  = 835.5
        self.params['anchor_radius']                        = 837.6
        self.params['mooring_diameter']                     = 0.0766

    def set_length_main(self, inval):
        self.params['main_section_height'] =  vecOption(inval/NSECTIONS, NSECTIONS)
        
    def set_length_aux(self, inval):
        self.params['offset_section_height'] =  vecOption(inval/NSECTIONS, NSECTIONS)
        
    def check_vectors(self):
        self.params['tower_outer_diameter']            = vecOption(self.params['tower_outer_diameter'], TOWSEC+1)
        self.params['tower_wall_thickness']            = vecOption(self.params['tower_wall_thickness'], TOWSEC)
        self.params['tower_section_height']            = vecOption(self.params['tower_section_height'], TOWSEC)
        self.params['main_outer_diameter']             = vecOption(self.params['main_outer_diameter'], NSECTIONS+1)
        self.params['main_wall_thickness']             = vecOption(self.params['main_wall_thickness'], NSECTIONS)
        self.params['main_stiffener_web_height']       = vecOption(self.params['main_stiffener_web_height'], NSECTIONS)
        self.params['main_stiffener_web_thickness']    = vecOption(self.params['main_stiffener_web_thickness'], NSECTIONS)
        self.params['main_stiffener_flange_width']     = vecOption(self.params['main_stiffener_flange_width'], NSECTIONS)
        self.params['main_stiffener_flange_thickness'] = vecOption(self.params['main_stiffener_flange_thickness'], NSECTIONS)
        self.params['main_stiffener_spacing']          = vecOption(self.params['main_stiffener_spacing'], NSECTIONS)
        self.params['main_bulkhead_thickness']         = vecOption(self.params['main_bulkhead_thickness'], NSECTIONS+1)
        #self.params['main_bulkhead_thickness'][:2] = 0.05
        
        self.params['offset_outer_diameter']             = vecOption(self.params['offset_outer_diameter'], NSECTIONS+1)
        self.params['offset_wall_thickness']             = vecOption(self.params['offset_wall_thickness'], NSECTIONS)
        self.params['offset_stiffener_web_height']       = vecOption(self.params['offset_stiffener_web_height'], NSECTIONS)
        self.params['offset_stiffener_web_thickness']    = vecOption(self.params['offset_stiffener_web_thickness'], NSECTIONS)
        self.params['offset_stiffener_flange_width']     = vecOption(self.params['offset_stiffener_flange_width'], NSECTIONS)
        self.params['offset_stiffener_flange_thickness'] = vecOption(self.params['offset_stiffener_flange_thickness'], NSECTIONS)
        self.params['offset_stiffener_spacing']          = vecOption(self.params['offset_stiffener_spacing'], NSECTIONS)
        self.params['offset_bulkhead_thickness']         = vecOption(self.params['offset_bulkhead_thickness'], NSECTIONS+1)
        #self.params['offset_bulkhead_thickness'][:2] = 0.05


    def set_reference(self, instr):
        if instr.upper() in Five_strings:

            self.params['hub_height']              = 90.0
            self.params['tower_outer_diameter']    = np.linspace(6.5, 3.87, TOWSEC+1)
            self.params['tower_section_height']    = vecOption(77.6/NSECTIONS, TOWSEC)
            self.params['tower_wall_thickness']    = np.linspace(0.027, 0.019, TOWSEC)
            self.params['Rhub']                 = 1.125
            self.params['main_freeboard'] = 10.0
            self.params['hub_cm']   = np.array([-5.01910, 0.0, 1.96256])

            if self.params.has_key('rna_mass'):
                self.params['rna_mass'] = 350e3 #285598.8
                self.params['rna_I'] = np.array([1.14930678e+08, 2.20354030e+07, 1.87597425e+07, 0.0, 5.03710467e+05, 0.0])
                self.params['rna_cg'] = np.array([-1.13197635, 0.0, 0.50875268])
                # Max thrust
                self.params['rna_force']      = np.array([ 1.03086517e+06, 0.0, -3.53463639e+06])
                self.params['rna_moment']     = np.array([9817509.35136043, 566675.9644231, -858920.77230378])
                # Max wind speed
                #self.params['rna_force']     = np.array([188038.8045, 0,  -16451.2637])
                #self.params['rna_moment']    = np.array([0.0, 131196.8431,  0.0])
                self.params['tip_position'] = np.array([-0.9315, 0.0, 27.5])
                
            
        elif instr.upper() in Ten_strings:

            #z Do Di Ixx= Iyy J
            #[m] [m] [m] [m4] [m4]
            dtuTowerData = StringIO("""
            145.63 5.50 5.44 2.03 4.07
            134.55 5.79 5.73 2.76 5.52
            124.04 6.07 6.00 3.63 7.26
            113.54 6.35 6.26 4.67 9.35
            103.03 6.63 6.53 5.90 11.80
            92.53 6.91 6.80 7.33 14.67
            82.02 7.19 7.07 8.99 17.98
            71.52 7.46 7.34 10.9 21.80
            61.01 7.74 7.61 13.08 26.15
            50.51 8.02 7.88 15.55 31.09
            40.00 8.30 8.16 15.55 31.09
            40.00 9.00 8.70 34.84 69.68
            38.00 9.00 8.70 35.74 71.48
            36.00 9.00 8.70 36.66 73.32
            34.00 9.00 8.70 37.59 75.18
            32.00 9.00 8.69 38.54 77.08
            30.00 9.00 8.69 39.51 79.01
            28.00 9.00 8.69 40.49 80.97
            26.00 9.00 8.69 41.48 82.97
            24.00 9.00 8.69 42.5 85.00
            22.00 9.00 8.69 43.53 87.05
            20.00 9.00 8.69 44.05 88.10
            16.00 9.00 8.80 28.09 56.18
            12.00 9.00 8.80 28.09 56.18
            8.00 9.00 8.80 28.09 56.18
            4.00 9.00 8.80 28.09 56.18
            0.00 9.00 8.80 28.09 56.18
            -8.400 9.00 8.80 28.09 56.18
            -16.800 9.00 8.80 28.09 56.18
            -25.20 9.00 8.80 28.09 56.18
            -33.60 9.00 8.80 28.09 56.18
            -42.60 9.00 8.80 28.09 56.18
            """)

            
            self.params['hub_height'] = 149.0
            self.params['main_freeboard'] = 30.0
            self.params['main_section_height'] += 3.0
            towerData = np.loadtxt(dtuTowerData)
            towerData = towerData[(towerData[:,0] >= 30.0),:]
            towerData = np.vstack((towerData[0,:], towerData))
            towerData[0,0] = self.params['hub_height']
            towerData[(towerData[:,0] == 40.0),0] += np.array([0.01, 0.0])
            trans_idx = np.where(towerData[:,1] == towerData[-1,1])[0]
            idx = [0, np.int_(trans_idx[0]/2), trans_idx[0], trans_idx[-1]]
            self.params['tower_section_height'] = np.diff( np.flipud( towerData[idx,0] ) )
            self.params['tower_outer_diameter'] = np.flipud( towerData[idx, 1] )
            t_temp = np.flipud( towerData[idx, 1] - towerData[idx, 2] )
            self.params['tower_wall_thickness'] = 0.5*(t_temp[:-1] + t_temp[1:])

            self.params['hub_cm']   = np.array([-7.1, 0.0, 2.75])

            if self.params.has_key('rna_mass'):
                self.params['rna_mass'] = 672300.5303006992
                self.params['rna_I'] = np.array([1.67424279e+08, 1.05417282e+08, 8.52142680e+07, 0.0, 1.53820004e+06, 0.0])
                self.params['rna_cg'] = np.array([-0.60484688, 0.0, 2.5177933 ])
                self.params['rna_force']   = np.array([ 2.11271060e+06, 0.0, -6.81485569e+06])
                self.params['rna_moment']  = np.array([29259007.24076359, 245729.82542406, -3075245.58067333])
                self.params['Rhub']     = 2.3
                self.params['tip_position'] = np.array([-3.551, 0.0, 21.504])
            
        else:
            raise ValueError('Inputs must be either NREL5MW or DTU10MW')

    def save(self, fname):
        assert type(fname) == type(''), 'Input filename must be a string'
        with open(fname,'wb') as fp:
            pickle.dump((self.params, self.design_variables, self.constraints, self.objective, self.optimizer), fp)
            
    def load(self, fname):
        assert type(fname) == type(''), 'Input filename must be a string'
        with open(fname,'rb') as fp:
            newobj = pickle.load(fp)

        if isinstance(newobj, tuple):
            newparams, newdesvar, newcons, newobj, newopt = newobj
            for k in newparams.keys():
                self.params[k] = newparams[k]

            if not newopt is None:
                self.set_optimizer(newopt)
            
            for k in newdesvar:
                if k[0].startswith('sm.'): k[0] = k[0][3:]
                self.add_design_variable(k[0], k[1], k[2])
            for k in newcons:
                if k[0].startswith('sm.'): k[0] = k[0][3:]
                self.add_constraint(k[0], k[1], k[2], k[3])

            if not newobj is None:
                self.add_objective(newobj[0], newobj[1])
            
        # Compatibility with older files
        elif isinstance(newobj, dict):
            for k in newobj.keys():
                self.params[k] = newobj[k]
        
    def get_assembly(self):
        return FloatingSE()

    
    def add_design_variable(self, varStr, lowVal, highVal):
        if self.optimizer is None:
            raise RuntimeError('Must set the optimizer to set the driver first')
        
        assert type(varStr) == type(''), 'Input variable must be a string'
        assert isinstance(lowVal, (float, int, np.float32, np.float64, np.int32, np.int64)), 'Invalid lower bound'
        assert isinstance(highVal, (float, int, np.float32, np.float64, np.int32, np.int64)), 'Invalid upper bound'
        
        if self.optimizer in ['COBLYA']:
            iscale=min(1.0 / lowVal, highVal / 1.0)
            self.prob.driver.add_desvar(varStr, lower=lowVal, upper=highVal, scaler=iscale)
        else:
            self.prob.driver.add_desvar(varStr, lower=lowVal, upper=highVal)

        # Internal record keeping
        self.design_variables.append([varStr, lowVal, highVal])

                                     
    def add_constraint(self, varStr, lowVal, highVal, eqVal):
        if self.optimizer is None:
            raise RuntimeError('Must set the optimizer to set the driver first')
        
        assert type(varStr) == type(''), 'Input variable must be a string'
        if not lowVal is None:
            assert isinstance(lowVal, (float, int, np.float32, np.float64, np.int32, np.int64)), 'Invalid lower bound'
        if not highVal is None:
            assert isinstance(highVal, (float, int, np.float32, np.float64, np.int32, np.int64)), 'Invalid upper bound'
        if not eqVal is None:
            assert isinstance(eqVal, (float, int, np.float32, np.float64, np.int32, np.int64)), 'Invalid equality value'

        self.prob.driver.add_constraint(varStr, lower=lowVal, upper=highVal, equals=eqVal)

        # Internal record keeping
        self.constraints.append([varStr, lowVal, highVal, eqVal])
        
            
    def set_optimizer(self, optStr):
        assert type(optStr) == type(''), 'Input optimizer must be a string'
        self.optimizer = optStr.upper()

        # Reset all design variables and constraints
        self.design_variables = []
        self.constraints = []
        self.objective = None
        
        # Establish the optimization driver
        if self.optimizer in ['SOGA','SOPSO','NM','SUBPLEX']:
            from openmdao.api import HeuristicDriverParallel
            self.prob.driver = HeuristicDriverParallel()
        #elif self.optimizer in ['SUBPLEX']:
            #from openmdao.api import HeuristicDriver
            #self.prob.driver = HeuristicDriver()
        elif self.optimizer in ['COBYLA','SLSQP']:
            self.prob.driver = ScipyOptimizer()
        elif self.optimizer in ['CONMIN', 'PSQP','SNOPT','NSGA2','ALPSO']:
            from openmdao.api import pyOptSparseDriver
            self.prob.driver = pyOptSparseDriver()
        else:
            raise ValueError('Unknown or unworking optimizer. '+self.optimizer)

        # Set default options
        self.prob.driver.options['optimizer'] = self.optimizer
        if self.optimizer == 'CONMIN':
            self.prob.driver.opt_settings['ITMAX'] = 1000
        elif self.optimizer in ['PSQP']:
            self.prob.driver.opt_settings['MIT'] = 1000
        elif self.optimizer in ['SOGA','SOPSO','NM','SUBPLEX']:
            self.prob.driver.options['population'] = 50
            self.prob.driver.options['generations'] = 500
        elif self.optimizer in ['NSGA2']:
            self.prob.driver.opt_settings['PopSize'] = 200
            self.prob.driver.opt_settings['maxGen'] = 500
        elif self.optimizer in ['SNOPT']:
            self.prob.driver.opt_settings['Iterations limit'] = 500
            self.prob.driver.opt_settings['Major optimality tolerance'] = 1e-4
            self.prob.driver.opt_settings['Major feasibility tolerance'] = 1e-6
            self.prob.driver.opt_settings['Minor feasibility tolerance'] = 1e-6
            self.prob.driver.opt_settings['Function precision'] = 1e-8
        elif self.optimizer in ['COBYLA','SLSQP']:
            self.prob.driver.options['tol'] = 1e-6
            self.prob.driver.options['maxiter'] = 1000


    def set_options(self, indict):
        if self.optimizer is None:
            raise RuntimeError('Must set the optimizer to set the driver first')
        assert isinstance(indict, dict), 'Options must be passed as a string:value dictionary'
        
        for k in indict.keys():
            if self.optimizer in ['SOGA','SOPSO','NM','SUBPLEX','COBYLA','SLSQP']:
                self.prob.driver.options[k] = indict[k]
            elif self.optimizer in ['CONMIN', 'PSQP','SNOPT','NSGA2','ALPSO']:
                if k in ['title','print_results','gradient method']:
                    self.prob.driver.options[k] = indict[k]
                else:
                    self.prob.driver.opt_settings[k] = indict[k]
            
    def get_constraints(self):

        conlist = [
            # Try to get tower height to match desired hub height
            ['tow.height_constraint', -1e-2, 1e-2, None],
            ['tow.tip_deflection_margin', 1.0, None, None],
            
            # Ensure that draft is greater than 0 (spar length>0) and that less than water depth
            ['main.draft_margin', None, 1.0, None],
            ['off.draft_margin', None, 1.0, None],
            ['main.wave_height_freeboard_ratio', None, 1.0, None],
            ['off.wave_height_freeboard_ratio', None, 1.0, None],
            ['wave_height_fairlead_ratio', None, 1.0, None],
            
            ['main_offset_spacing', 1.0, None, None],
            
            # Ensure that the radius doesn't change dramatically over a section
            ['main.manufacturability', 0.0, None, None],
            ['main.weldability', None, 0.0, None],
            ['off.manufacturability', 0.0, None, None],
            ['off.weldability', None, 0.0, None],
            ['tow.manufacturability', 0.0, None, None],
            ['tow.weldability', None, 0.0, None],
            
            # Ensure that the spar top matches the tower main
            ['tower_transition_buffer', 0.0, None, None],
            ['nacelle_transition_buffer', 0.0, None, None],

            # Make sure semisub columns don't get submerged
            ['offset_freeboard_heel_margin', 0.0, None, None],
            ['offset_draft_heel_margin', 0.0, None, None],
            
            # Ensure max mooring line tension is less than X% of MBL: 60% for intact mooring, 80% for damanged
            ['axial_unity', 0.0, 1.0, None],
            
            # Ensure there is sufficient mooring line length, MAP doesn't give an error about this
            ['mooring_length_max', None, 1.0, None],
            
            # API Bulletin 2U constraints
            ['main.flange_spacing_ratio', None, 1.0, None],
            ['main.stiffener_radius_ratio', None, 0.5, None],
            ['main.flange_compactness', 1.0, None, None],
            ['main.web_compactness', 1.0, None, None],
            ['main.axial_local_api', None, 1.0, None],
            ['main.axial_general_api', None, 1.0, None],
            ['main.external_local_api', None, 1.0, None],
            ['main.external_general_api', None, 1.0, None],
            
            ['off.flange_spacing_ratio', None, 1.0, None],
            ['off.stiffener_radius_ratio', None, 0.5, None],
            ['off.flange_compactness', 1.0, None, None],
            ['off.web_compactness', 1.0, None, None],
            ['off.axial_local_api', None, 1.0, None],
            ['off.axial_general_api', None, 1.0, None],
            ['off.external_local_api', None, 1.0, None],
            ['off.external_general_api', None, 1.0, None],
            
            # Pontoon tube radii
            #['main_connection_ratio', 0.0, None, None],
            #['offset_connection_ratio', 0.0, None, None],
            
            # Pontoon stress safety factor
            ['pontoon_stress', None, 1.0, None],
            ['pontoon_wave_height_depth_margin', 1.0, None, None],

            # Frame3DD estimates of stress and buckling
            ['tower_stress', None, 1.0, None],
            ['tower_shell_buckling', None, 1.0, None],
            ['tower_global_buckling', None, 1.0, None],

            ['main_stress', None, 1.0, None],
            ['main_shell_buckling', None, 1.0, None],
            ['main_global_buckling', None, 1.0, None],

            ['offset_stress', None, 1.0, None],
            ['offset_shell_buckling', None, 1.0, None],
            ['offset_global_buckling', None, 1.0, None],
            
            # Achieving non-zero variable ballast height means the semi can be balanced with margin as conditions change
            ['variable_ballast_height_ratio', 0.0, 1.0, None],
            ['variable_ballast_mass', 0.0, None, None],
            
            # Metacentric height should be positive for static stability
            ['metacentric_height', 1.0, None, None],
            
            # Center of buoyancy should be above CG (difference should be positive, None],
            #['buoyancy_to_gravity', 0.1, None, None],
            
            # Surge restoring force should be greater than wave-wind forces (ratio < 1, None],
            ['offset_force_ratio', None, 1.0, None],
            
            # Heel angle should be less than 6deg for ordinary operation, less than 10 for extreme conditions
            ['heel_moment_ratio', None, 1.0, None],

            # Wave forcing period should be different than natural periods and structural modes
            ['period_margin_low', None, 1.0, None],
            ['period_margin_high', 1.0, None, None],
            ['modal_margin_low', None, 1.0, None],
            ['modal_margin_high', 1.0, None, None]
        ]
        #raise NotImplementedError("Subclasses should implement this!")
        return self.constraints if len(self.constraints) > 0 else conlist


    def constraint_report(self, tol=2e-2, printFlag=True):
        passStr   = 'yes'
        noStr     = 'NO'
        activeStr = 'ACTIVE'
        eps = 1e-12
        
        deslist = self.design_variables
        for k in deslist:
            myval   = self.params[k[0]]
            normLin = (myval - k[1]) / (k[2] - k[1])
            normLog = (np.log(myval+eps) - np.log(k[1]+eps)) / (np.log(k[2]+eps) - np.log(k[1]+eps))
            margin  = np.maximum(np.minimum(normLin, 1.0-normLin), np.minimum(normLog, 1.0-normLog))
            lowStr  = str(k[1])+' <\t'
            highStr = '< '+str(k[2])+'\t'
            conStr  = activeStr if np.any(margin <= tol) else ''
            valStr  = str(self.params[k[0]])
            if printFlag: print(conStr, lowStr, k[0], highStr, valStr)
        
        #print('Status\tLow\tName\tHigh\tEq\tValue')
        conlist = self.constraints if len(self.constraints) > 0 else self.get_constraints()
        allPass = True
        for k in conlist:
            lowStr   = ''
            highStr  = ''
            eqStr    = ''
            passFlag = True
            myval    = self.prob[k[0]]
            margin   = 0.0 if type(myval) == type(0.0) else np.inf * np.ones(myval.shape)
            if not k[1] is None:
                lowStr   = str(k[1])+' <\t'
                passFlag = passFlag and np.all(myval >= k[1])
                normVal  = myval - k[1]
                if k[1] != 0.0: normVal /= k[1]
                margin = np.minimum(margin, np.abs(normVal))
                
            if not k[2] is None:
                highStr  = '< '+str(k[2])+'\t'
                passFlag = passFlag and np.all(myval <= k[2])
                normVal  = k[2] - myval
                if k[2] != 0.0: normVal /= k[2]
                margin = np.minimum(margin, np.abs(normVal))

            if not k[3] is None:
                highStr = '= '+str(k[3])+'\t'
                passFlag = passFlag and np.all(myval == k[3])
                normVal  = np.abs(k[3] - myval)
                if k[3] != 0.0: normVal /= k[3]
                margin = np.minimum(margin, np.abs(normVal))

            valStr = str(myval)
            if not passFlag:
                conStr = noStr+'-'+activeStr
            elif np.any(margin <= tol):
                conStr = activeStr
            else:
                conStr = passStr
                valStr = ''
            if printFlag: print(conStr, '\t', lowStr, k[0], '\t', highStr, eqStr, '\t', valStr)

            # Record summary statistic
            allPass = allPass and passFlag
            
        return allPass
    
            
    def add_objective(self, varname='total_cost', scale=1e-9):
        if (len(self.prob.driver._objs) == 0):
            self.prob.driver.add_objective(varname, scaler=scale)
            
        self.objective = (varname, scale)

    def set_inputs(self):
        # Load all variables from local params dictionary
        localnames = self.params.keys()

        for ivar in localnames:
            try:
                self.prob[ivar] = self.params[ivar]
            except KeyError:
                print('Cannot set: ', ivar, '=', self.params[ivar])
                continue
            except AttributeError as e:
                print('Vector issues?: ', ivar)
                print(e)
                raise e
            except ValueError as e:
                print('Badding setting of: ', ivar, '=', self.params[ivar])
                print(e)
                raise e

        # Check that everything got set correctly
        namesAssembly = self.prob.root._unknowns_dict.keys()
        for ivar in namesAssembly:
            if self.prob.root._unknowns_dict[ivar].has_key('_canset_') and self.prob.root._unknowns_dict[ivar]['_canset_']:
                selfvar = ivar.split('.')[-1]
                if not selfvar in localnames:
                    print('WARNING:', selfvar, 'has not been set!')
                    

        '''
        namesAssembly = self.prob.root._unknowns_dict.keys()
        for ivar in namesAssembly:
            if self.prob.root._unknowns_dict[ivar].has_key('_canset_') and self.prob.root._unknowns_dict[ivar]['_canset_']:
                selfvar = ivar.split('.')[-1]

                selfval = self.prob[selfvar]
                if ( (type(selfval) == type(0.0)) or (type(selfval) == np.float64) or (type(selfval) == np.float32) ) and selfval == 0.0:
                    print(selfvar, 'is zero! Did you set it?')
                if ( (type(selfval) == type(0)) or (type(selfval) == np.int64) or (type(selfval) == np.int32) ) and selfval == 0:
                    print(selfvar, 'is zero! Did you set it?')
                elif type(selfval) == type(np.array([])) and not np.any(selfval):
                    print(selfvar, 'is zero! Did you set it?')

                selfval = getattr(self, selfvar, None)
                if selfval is None:
                    print('Variable not found:', ivar, selfvar, self.prob[ivar])
                else:
                    self.prob[ivar] = selfval
        #raise NotImplementedError("Subclasses should implement this!")
        '''

    def store_results(self):
        localnames = self.params.keys()
        optDict = self.prob.driver.get_desvars()
        for ivar in optDict.keys():
            ival = optDict[ivar]
            if type(ival) == type(np.array([])) and len(ival) == 1: ival=ival[0]
            selfvar = ivar.split('.')[-1]
            self.prob[ivar] = ival
            if selfvar in localnames:
                self.params[ivar] = ival
                ivalstr = 'np.array( '+str(ival.tolist())+' )' if type(ival)==type(np.array([])) else str(ival)
                print(ivar, '=', ivalstr)

                
    def init_problem(self, optFlag=False):
        self.prob.root = self.get_assembly()
        
        if optFlag:
            if self.optimizer is None:
                raise RuntimeError('Must set the optimizer to set the driver first')
            if len(self.design_variables) == 0:
                raise RuntimeError('Must set design variables before running optimization')
            if len(self.constraints) == 0:
                print('Warning: no constraints set')
            if self.objective is None:
                self.add_objective()

        # Recorder
        #recorder = DumpRecorder('floatingOptimization.dat')
        #recorder.options['record_params'] = True
        #recorder.options['record_metadata'] = False
        #recorder.options['record_derivs'] = False
        #self.prob.driver.add_recorder(recorder)
        
        # Note this command must be done after the constraints, design variables, and objective have been set,
        # but before the initial conditions are specified (unless we use the default initial conditions )
        # After setting the intial conditions, running setup() again will revert them back to default values
        self.prob.setup()

        self.set_inputs()

        # Checks
        self.prob.check_setup()
        #self.prob.check_total_derivatives()

    def run(self):
        if self.optimizer is None:
            print('WARNING: executing once because optimizer is not set')
            self.evaluate()
        else:
            self.init_problem(optFlag=True)
            self.prob.run()
            self.store_results()
            print(self.prob.driver.get_constraints())
            #print(self.prob.driver.get_desvars())
            print(self.prob.driver.get_objectives())
        
    def evaluate(self):
        self.init_problem(optFlag=False)
        self.prob.run_once()


    def visualize(self, fname=None):
        raise NotImplementedError("Subclasses should implement this!")

    def init_figure(self):
        from mayavi import mlab
        mysky   = np.array([135, 206, 250]) / 255.0
        mysky   = tuple(mysky.tolist())
        #fig = plt.figure()
        #ax = fig.add_subplot(111, projection='3d')
        #fig = mlab.figure(bgcolor=(1,)*3, size=(1600,1100))
        #fig = mlab.figure(bgcolor=mysky, size=(1600,1100))
        fig = mlab.figure(bgcolor=(0,)*3, size=(1600,1100))
        return fig

    def draw_ocean(self, fig=None):
        from mayavi import mlab
        if fig is None: fig=self.init_figure()
        npts = 100
        
        #mybrown = np.array([244, 170, 66]) / 255.0
        #mybrown = tuple(mybrown.tolist())
        mywater = np.array([95, 158, 160 ]) / 255.0 #(0.0, 0.0, 0.8) [143, 188, 143]
        mywater = tuple(mywater.tolist())
        alpha   = 0.3

        # Waterplane box
        x = y = 100 * np.linspace(-1, 1, npts)
        X,Y = np.meshgrid(x,y)
        Z   = np.sin(100*X*Y) #np.zeros(X.shape)
        #ax.plot_surface(X, Y, Z, alpha=alpha, color=mywater)
        mlab.mesh(X, Y, Z, opacity=alpha, color=mywater, figure=fig)
        
        # Sea floor
        Z = -self.params['water_depth'] * np.ones(X.shape)
        #ax.plot_surface(10*X, 10*Y, Z, alpha=1.0, color=mybrown)
        #mlab.mesh(10*X,10*Y,Z, opacity=1.0, color=mybrown, figure=fig)

        # Sides
        #x = 500 * np.linspace(-1, 1, npts)
        #z = self.params['water_depth'] * np.linspace(-1, 0, npts)
        #X,Z = np.meshgrid(x,z)
        #Y = x.max()*np.ones(Z.shape)
        ##ax.plot_surface(X, Y, Z, alpha=alpha, color=mywater)
        #mlab.mesh(X,Y,Z, opacity=alpha, color=mywater, figure=fig)
        #mlab.mesh(X,-Y,Z, opacity=alpha, color=mywater, figure=fig)
        #mlab.mesh(Y,X,Z, opacity=alpha, color=mywater, figure=fig)
        ##mlab.mesh(-Y,X,Z, opacity=alpha, color=mywater, figure=fig)

    def draw_mooring(self, fig, mooring):
        from mayavi import mlab
        mybrown = np.array([244, 170, 66]) / 255.0
        mybrown = tuple(mybrown.tolist())
        npts    = 100
        
        # Sea floor
        r  = np.linspace(0, self.params['anchor_radius'], npts)
        th = np.linspace(0, 2*np.pi, npts)
        R, TH = np.meshgrid(r, th)
        X = R*np.cos(TH)
        Y = R*np.sin(TH)
        Z = -self.params['water_depth'] * np.ones(X.shape)
        #ax.plot_surface(X, Y, Z, alpha=1.0, color=mybrown)
        mlab.mesh(X,Y,Z, opacity=1.0, color=mybrown, figure=fig)
        cmoor = (0,0.8,0)
        nlines = int( self.params['number_of_mooring_connections'] * self.params['mooring_lines_per_connection'] )
        for k in xrange(nlines):
            #ax.plot(mooring[k,:,0], mooring[k,:,1], mooring[k,:,2], 'k', lw=2)
            mlab.plot3d(mooring[k,:,0], mooring[k,:,1], mooring[k,:,2], color=cmoor, tube_radius=0.5*self.params['mooring_diameter'], figure=fig)

            
    def draw_pontoons(self, fig, truss, R, freeboard):
        from mayavi import mlab
        nE = truss.shape[0]
        c = (0.5,0,0)
        for k in xrange(nE):
            if np.any(truss[k,2,:] > freeboard): continue
            mlab.plot3d(truss[k,0,:], truss[k,1,:], truss[k,2,:], color=c, tube_radius=R, figure=fig)

            
    def draw_column(self, fig, centerline, freeboard, h_section, r_nodes, spacingVec=None, ckIn=None):
        from mayavi import mlab
        npts = 20

        nsection = h_section.size
        z_nodes = np.flipud( freeboard - np.r_[0.0, np.cumsum(np.flipud(h_section))] )

        th = np.linspace(0, 2*np.pi, npts)
        for k in xrange(nsection):
            rk = np.linspace(r_nodes[k], r_nodes[k+1], npts)
            z  = np.linspace(z_nodes[k], z_nodes[k+1], npts)
            R, TH = np.meshgrid(rk, th)
            Z, _  = np.meshgrid(z, th)
            X = R*np.cos(TH) + centerline[0]
            Y = R*np.sin(TH) + centerline[1]

            # Draw parameters
            if ckIn is None:
                ck = (0.6,)*3 if np.mod(k,2) == 0 else (0.4,)*3
            else:
                ck = ckIn
            #ax.plot_surface(X, Y, Z, alpha=0.5, color=ck)
            mlab.mesh(X, Y, Z, opacity=0.7, color=ck, figure=fig)

            if spacingVec is None: continue
            
            z = z_nodes[k] + spacingVec[k]
            while z < z_nodes[k+1]:
                rk = np.interp(z, z_nodes[k:], r_nodes[k:])
                #ax.plot(rk*np.cos(th), rk*np.sin(th), z*np.ones(th.shape), 'r', lw=0.25)
                mlab.plot3d(rk*np.cos(th) + centerline[0], rk*np.sin(th) + centerline[1], z*np.ones(th.shape), color=(0.5,0,0), figure=fig)
                z += spacingVec[k]
                
                '''
                # Web
                r   = np.linspace(rk - self.params['stiffener_web_height'][k], rk, npts)
                R, TH = np.meshgrid(r, th)
                Z, _  = np.meshgrid(z, th)
                X = R*np.cos(TH)
                Y = R*np.sin(TH)
                ax.plot_surface(X, Y, Z, alpha=0.7, color='r')

                # Flange
                r = r[0]
                h = np.linspace(0, self.params['stiffener_flange_width'][k], npts)
                zflange = z + h - 0.5*self.params['stiffener_flange_width'][k]
                R, TH = np.meshgrid(r, th)
                Z, _  = np.meshgrid(zflange, th)
                X = R*np.cos(TH)
                Y = R*np.sin(TH)
                ax.plot_surface(X, Y, Z, alpha=0.7, color='r')
                '''

    def draw_ballast(self, fig, centerline, freeboard, h_section, r_nodes, h_perm, h_water):
        from mayavi import mlab
        npts = 40
        th = np.linspace(0, 2*np.pi, npts)
        z_nodes = np.flipud( freeboard - np.r_[0.0, np.cumsum(np.flipud(h_section))] )

        # Permanent ballast
        z_perm = z_nodes[0] + np.linspace(0, h_perm, npts)
        r_perm = np.interp(z_perm, z_nodes, r_nodes)
        R, TH = np.meshgrid(r_perm, th)
        Z, _  = np.meshgrid(z_perm, th)
        X = R*np.cos(TH) + centerline[0]
        Y = R*np.sin(TH) + centerline[1]
        ck = np.array([122, 85, 33]) / 255.0
        ck = tuple(ck.tolist())
        mlab.mesh(X, Y, Z, color=ck, figure=fig)

        # Water ballast
        z_water = z_perm[-1] + np.linspace(0, h_water, npts)
        r_water = np.interp(z_water, z_nodes, r_nodes)
        R, TH = np.meshgrid(r_water, th)
        Z, _  = np.meshgrid(z_water, th)
        X = R*np.cos(TH) + centerline[0]
        Y = R*np.sin(TH) + centerline[1]
        ck = (0.0, 0.1, 0.8) # Dark blue
        mlab.mesh(X, Y, Z, color=ck, figure=fig)

    def draw_buoyancy_tank(self, fig, centerline, freeboard, h_section, loc, r_box, h_box):
        from mayavi import mlab
        npts = 20
        z_nodes = np.flipud( freeboard - np.r_[0.0, np.cumsum(np.flipud(h_section))] )
        z_lower = loc*(z_nodes[-1] - z_nodes[0]) + z_nodes[0]

        # Lower and Upper surfaces
        r  = np.linspace(0, r_box, npts)
        th = np.linspace(0, 2*np.pi, npts)
        R, TH = np.meshgrid(r, th)
        X = R*np.cos(TH) + centerline[0]
        Y = R*np.sin(TH) + centerline[1]
        Z = z_lower * np.ones(X.shape)
        ck = (0.9,)*3
        mlab.mesh(X,Y,Z, opacity=0.5, color=ck, figure=fig)
        Z += h_box
        mlab.mesh(X,Y,Z, opacity=0.5, color=ck, figure=fig)

        # Cylinder part
        z  = z_lower + np.linspace(0, h_box, npts)
        Z, TH  = np.meshgrid(z, th)
        R = r_box * np.ones(Z.shape)
        X = R*np.cos(TH) + centerline[0]
        Y = R*np.sin(TH) + centerline[1]
        mlab.mesh(X, Y, Z, opacity=0.5, color=ck, figure=fig)
        
    def set_figure(self, fig, fname=None):
        from mayavi import mlab
        #ax.set_aspect('equal')
        #set_axes_equal(ax)
        #ax.autoscale_view(tight=True)
        #ax.set_xlim([-125, 125])
        #ax.set_ylim([-125, 125])
        #ax.set_zlim([-220, 30])
        #plt.axis('off')
        #plt.show()
        #mlab.move([-517.16728532, -87.0711504, 5.60826224], [1.35691603e+01, -2.84217094e-14, -1.06547500e+02])
        #mlab.view(-170.68320804213343, 78.220729198686854, 549.40101471336777, [1.35691603e+01,  0.0, -1.06547500e+02])
        if not fname is None: mlab.savefig(fname, figure=fig)
        mlab.show(stop=True)

        
    



"""PMSG_arms.py
Created by Latha Sethuraman, Katherine Dykes.
Copyright (c) NREL. All rights reserved.
Electromagnetic design based on conventional magnetic circuit laws
Structural design based on McDonald's thesis """

from openmdao.api import Group, Problem, ExplicitComponent,ExecComp,IndepVarComp,ScipyOptimizeDriver
import numpy as np
from math import pi, cos, sqrt, radians, sin, exp, log10, log, tan, atan

class PMSG_Arms(ExplicitComponent):
    """ Estimates overall mass dimensions and Efficiency of PMSG -arms generator. """
    
    def setup(self):
        
        # PMSG_arms generator design inputs
        #self.add_input('r_s', val=0.0, units ='m', desc='airgap radius r_s')
        self.add_input('rad_ag', val=0.0, units ='m', desc='airgap radius')
        self.add_input('l_s', val=0.0, units ='m', desc='Stator core length l_s')
        self.add_input('h_s', val=0.0, units ='m', desc='Yoke height h_s')
        self.add_input('tau_p', val=0.0, units ='m', desc='Pole pitch self.tau_p')
        self.add_input('machine_rating', val=0.0, units ='W', desc='Machine rating')
        self.add_input('n_nom', val=0.0, units ='rpm', desc='rated speed')
        self.add_input('Torque', val=0.0, units ='N*m', desc='Rated torque ')
        self.add_input('h_m', val=0.0, units ='m', desc='magnet height')
        self.add_input('h_ys', val=0.0, units ='m', desc='Yoke height')
        self.add_input('h_yr', val=0.0, units ='m', desc='rotor yoke height')
        
        # structural design variables
        self.add_input('n_s', val=0.0, desc='number of stator arms n_s')
        self.add_input('b_st', val=0.0, units ='m', desc='arm width b_st')
        self.add_input('d_s', val=0.0,units ='m', desc='arm depth d_s')
        self.add_input('t_ws', val=0.0,units ='m', desc='arm depth thickness self.t_wr')
        self.add_input('n_r', val=0.0, desc='number of arms n')
        self.add_input('b_r', val=0.0,units ='m', desc='arm width b_r')
        self.add_input('d_r', val=0.0, units ='m', desc='arm depth d_r')
        self.add_input('t_wr', val=0.0, units ='m', desc='arm depth thickness self.t_wr')
        self.add_input('R_o', val=0.0, units ='m',desc='Shaft radius')
        
        # PMSG_arms generator design outputs
        
        # Magnetic loading
        self.add_output('B_symax', val=0.0, desc='Peak Stator Yoke flux density B_ymax')
        self.add_output('B_tmax', val=0.0, desc='Peak Teeth flux density')
        self.add_output('B_rymax', val=0.0, desc='Peak Rotor yoke flux density')
        self.add_output('B_smax', val=0.0, desc='Peak Stator flux density')
        self.add_output('B_pm1', val=0.0, desc='Fundamental component of peak air gap flux density')
        self.add_output('B_g', val=0.0, desc='Peak air gap flux density B_g')
        
        # Stator design
        self.add_output('N_s', val=0.0, desc='Number of turns in the stator winding')
        self.add_output('b_s', val=0.0, desc='slot width')
        self.add_output('b_t', val=0.0, desc='tooth width')
        self.add_output('A_Cuscalc', val=0.0, desc='Conductor cross-section mm^2')
        self.add_output('S'       , val=0.0, desc='Stator slots')
        
        # Rotor magnet dimension
        self.add_output('b_m', val=0.0, desc='magnet width')
        self.add_output('p', val=0.0, desc='No of pole pairs')
        
        # Electrical performance
        self.add_output('E_p', val=0.0, desc='Stator phase voltage')
        self.add_output('f', val=0.0, desc='Generator output frequency')
        self.add_output('I_s', val=0.0, desc='Generator output phase current')
        self.add_output('R_s', val=0.0, desc='Stator resistance')
        self.add_output('L_s', val=0.0, desc='Stator synchronising inductance')
        self.add_output('A_1', val=0.0, desc='Electrical loading')
        self.add_output('J_s', val=0.0, desc='Current density')
        
        # Objective functions
        self.add_output('Mass', val=0.0, desc='Actual mass')
        self.add_output('K_rad', val=0.0, desc='K_rad')
        self.add_output('Losses', val=0.0, desc='Total loss')
        self.add_output('gen_eff', val=0.0, desc='Generator efficiency')
        
        # Structural performance
        self.add_output('u_Ar', val=0.0, desc='Rotor radial deflection')
        self.add_output('y_Ar', val=0.0, desc='Rotor axial deflection')
        self.add_output('z_A_r', val=0.0, desc='Rotor circumferential deflection')      
        self.add_output('u_As', val=0.0, desc='Stator radial deflection')
        self.add_output('y_As', val=0.0, desc='Stator axial deflection')
        self.add_output('z_A_s', val=0.0, desc='Stator circumferential deflection')  
        self.add_output('u_all_r', val=0.0, desc='Allowable radial rotor')
        self.add_output('u_all_s', val=0.0, desc='Allowable radial stator')
        self.add_output('y_all', val=0.0, desc='Allowable axial')
        self.add_output('z_all_s', val=0.0, desc='Allowable circum stator')
        self.add_output('z_all_r', val=0.0, desc='Allowable circum rotor')
        self.add_output('b_all_s', val=0.0, desc='Allowable arm')
        self.add_output('b_all_r', val=0.0, desc='Allowable arm dimensions')
        self.add_output('TC1', val=0.0, desc='Torque constraint')
        self.add_output('TC2', val=0.0, desc='Torque constraint-rotor')
        self.add_output('TC3', val=0.0, desc='Torque constraint-stator')
        
        # Other parameters
        self.add_output('R_out', val=0.0, desc='Outer radius')
        
        self.add_output('Slot_aspect_ratio', val=0.0, desc='Slot aspect ratio')
        
        # Mass Outputs
        self.add_output('mass_PM', val=0.0, units='kg', desc='Magnet mass')
        self.add_output('Copper', val=0.0, units='kg', desc='Copper Mass')
        self.add_output('Iron', val=0.0, units='kg', desc='Electrical Steel Mass')
        self.add_output('Structural_mass', val=0.0, units='kg', desc='Structural Mass')
        
        # Material properties
        self.add_input('rho_Fes', val=0.0,units='kg*m**-3', desc='Structural Steel density ')
        self.add_input('rho_Fe', val=0.0,units='kg*m**-3', desc='Magnetic Steel density ')
        self.add_input('rho_Copper', val=0.0,units='kg*m**-3', desc='Copper density ')
        self.add_input('rho_PM', val=0.0,units='kg*m**-3', desc='Magnet density ')
        
        # inputs/outputs for interface with drivese    
        self.add_input('shaft_cm', val= np.zeros(3), units='m', desc='Main Shaft CM')
        self.add_input('shaft_length', val=0.0, units='m', desc='main shaft length')
        self.add_output('I', val=np.zeros(3),desc='Moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass')
        self.add_output('cm', val=np.zeros(3),desc='COM [x,y,z]')
        
        #self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

        
    def compute(self, inputs, outputs):
        # Unpack inputs
        #r_s               = inputs['r_s']
        rad_ag            = inputs['rad_ag']
        l_s               = inputs['l_s']
        h_s               = inputs['h_s']
        tau_p             = inputs['tau_p']
        h_m               = inputs['h_m']
        h_ys              = inputs['h_ys']
        h_yr              = inputs['h_yr']
        machine_rating    = inputs['machine_rating']
        n_nom             = inputs['n_nom']
        Torque            = inputs['Torque']
        
        b_st              = inputs['b_st']
        d_s               = inputs['d_s']
        t_ws              = inputs['t_ws']
        n_r               = inputs['n_r']
        n_s               = inputs['n_s']
        b_r               = inputs['b_r']
        d_r               = inputs['d_r']
        t_wr              = inputs['t_wr']
    
        R_o               = inputs['R_o']
        rho_Fe            = inputs['rho_Fe']
        rho_Copper        = inputs['rho_Copper']
        rho_Fes           = inputs['rho_Fes']
        rho_PM            = inputs['rho_PM']
        shaft_cm          = inputs['shaft_cm']
        shaft_length      = inputs['shaft_length']

        
        # Assign values to universal constants
        B_r    = 1.2                 # Tesla remnant flux density
        g1     = 9.81                # m / s^2 acceleration due to gravity
        E      = 2e11                # N / m^2 young's modulus
        sigma  = 40e3                # shear stress assumed
        ratio  = 0.7                 # ratio of magnet width to pole pitch(bm / tau_p)
        mu_0   = pi * 4e-7           # permeability of free space
        mu_r   = 1.06                # relative permeability 
        phi    = radians(90)         # tilt angle (rotor tilt -90 degrees during transportation)
        cofi   = 0.85                # power factor
        
        # Assign values to design constants
        h_w       = 0.005            # Slot wedge height
        h_i       = 0.001            # coil insulation thickness
        y_tau_p   = 1                # Coil span to pole pitch
        m         = 3                # no of phases
        q1        = 1                # no of slots per pole per phase
        b_s_tau_s = 0.45             # slot width to slot pitch ratio
        k_sfil    = 0.65             # Slot fill factor
        P_Fe0h    = 4                # specific hysteresis losses W / kg @ 1.5 T 
        P_Fe0e    = 1                # specific eddy losses W / kg @ 1.5 T 
        rho_Cu    = 1.8e-8 * 1.4     # Copper resisitivty
        k_fes     = 0.9              # Stator iron fill factor per Grauers
        b_so      =  0.004           # Slot opening
        alpha_p   =  pi / 2 * 0.7

        # back iron thickness for rotor and stator
        t_s = h_ys    
        t   = h_yr
        
        
        ###################################################### Electromagnetic design#############################################
        
        K_rad = l_s / (2 * rad_ag)           # Aspect ratio
        #T     = Torque                    # rated torque
        l_u   = k_fes * l_s               # useful iron stack length
        We    = tau_p
        l_b   = 2 * tau_p                 # end winding length
        l_e   = l_s + 2 * 0.001 * rad_ag     # equivalent core length
        b_m   = 0.7 * tau_p               # magnet width     
        
        
        # Calculating air gap length
        dia   =  2 * rad_ag                  # air gap diameter
        ag_len     =  0.001 * dia              # air gap length
        r_m   =  rad_ag + h_ys + h_s         # magnet radius
        r_r   =  rad_ag - ag_len                  # rotor radius
        
        p     =  np.round(pi * dia / (2 * tau_p))  # pole pairs
        f     =  n_nom * p / 60                    # outout frequency
        S     = 2 * p*q1 * m                       # Stator slots
        N_conductors = S * 2                    
        N_s = N_conductors / 2/3                   # Stator turns per phase
        tau_s = pi * dia / S                       # Stator slot pitch
        b_s    =  b_s_tau_s * tau_s                # slot width 
        b_t    =  tau_s - (b_s)                    # tooth width
        Slot_aspect_ratio = h_s / b_s
        
        # Calculating Carter factor for statorand effective air gap length
        gamma  =  4 / pi * (b_so / 2/(ag_len + h_m / mu_r) * atan(b_so / 2/(ag_len + h_m / mu_r)) - log(sqrt(1 + (b_so / 2/(ag_len + h_m / mu_r))**2)))
        k_C    =  tau_s / (tau_s - gamma * (ag_len + h_m / mu_r))   # carter coefficient
        g_eff  =  k_C * (ag_len + h_m / mu_r)
        
        # angular frequency in radians
        om_m            =  2 * pi * n_nom / 60
        om_e            =  p * om_m / 2

        
        # Calculating magnetic loading
        B_pm1   = B_r * h_m / mu_r / g_eff
        B_g     = B_r * h_m / mu_r / g_eff * (4 / pi) * sin(alpha_p)
        B_symax = B_g * b_m * l_e / (2 * h_ys * l_u)
        B_rymax = B_g * b_m * l_e / (2 * h_yr * l_s)
        B_tmax  = B_g * tau_s / b_t
        
        # Calculating winding factor
        k_wd    = sin(pi / 6) / q1 / sin(pi / 6/q1)      
        
        L_t = l_s + 2 * tau_p
        
        #l = L_t                          # length - now using L_t everywhere
        
        # Stator winding length, cross-section and resistance
        l_Cus     = 2 * N_s * (2 * tau_p + L_t)
        A_s       = b_s        * (h_s - h_w)        * q1 * p
        A_scalc   = b_s * 1000 * (h_s - h_w) * 1000 * q1 * p
        A_Cus     = A_s * k_sfil / N_s
        A_Cuscalc = A_scalc * k_sfil / N_s
        R_s       = l_Cus * rho_Cu / A_Cus
        
        # Calculating leakage inductance in  stator
        L_m        = 2 * mu_0 * N_s**2 / p * m * k_wd**2 * tau_p * L_t / pi**2 / g_eff
        L_ssigmas  = 2 * mu_0 * N_s**2 / p / q1 * l_s  * ((h_s - h_w) / (3 * b_s) + h_w / b_so)                        # slot leakage inductance
        L_ssigmaew = 2 * mu_0 * N_s**2 / p / q1 * l_s  * 0.34 * ag_len * (l_e - 0.64 * tau_p * y_tau_p) / l_s          # end winding leakage inductance
        L_ssigmag  = 2 * mu_0 * N_s**2 / p / q1 * l_s  * (5 * (ag_len * k_C / b_so) / (5 + 4 * (ag_len * k_C / b_so))) # tooth tip leakage inductance
        L_ssigma   = L_ssigmas + L_ssigmaew + L_ssigmag
        L_s  = L_m + L_ssigma
        
        # Calculating no-load voltage induced in the stator
        E_p    = 2 * N_s * L_t * rad_ag * k_wd * om_m * B_g / sqrt(2)
        
        Z = machine_rating / (m * E_p)        
        G = E_p**2 - (om_e * L_s * Z)**2

        # Calculating stator current and electrical loading
        I_s     = sqrt(Z**2 + (((E_p - G**0.5) / (om_e * L_s)**2)**2))
        J_s     = I_s / A_Cuscalc
        A_1     = 6 * N_s * I_s / (pi * dia)
        I_snom  = machine_rating / (m * E_p * cofi) # rated current                
        I_qnom  = machine_rating / (m * E_p)
        X_snom  = om_e * (L_m + L_ssigma)
                
        B_smax  = sqrt(2) * I_s * mu_0 / g_eff
        
        # Calculating Electromagnetically active mass
        
        V_Cus   = m * l_Cus * A_Cus              # copper volume
        V_Fest  = L_t * 2*p * q1 * m * b_t * h_s # volume of iron in stator tooth
        
        V_Fesy  = L_t * pi * ((rad_ag + h_s + h_ys)**2 - (rad_ag + h_s)**2) # volume of iron in stator yoke
        V_Fery  = L_t * pi * ((r_r - h_m)**2 - (r_r - h_m - h_yr)**2)
        Copper  = V_Cus * rho_Copper

        M_Fest  = V_Fest * rho_Fe    # Mass of stator tooth
        M_Fesy  = V_Fesy * rho_Fe    # Mass of stator yoke
        M_Fery  = V_Fery * rho_Fe    # Mass of rotor yoke
        Iron    = M_Fest + M_Fesy + M_Fery
        
        # Calculating Losses
        ##1. Copper Losses
        
        K_R = 1.2   # Skin effect correction co - efficient
        P_Cu        = m * I_snom**2 * R_s * K_R

        # Iron Losses ( from Hysteresis and eddy currents) 
        P_Hyys    = M_Fesy * (B_symax / 1.5)**2 * (P_Fe0h * om_e / (2 * pi * 60))      # Hysteresis losses in stator yoke
        P_Ftys    = M_Fesy * (B_symax / 1.5)**2 * (P_Fe0e * (om_e / (2 * pi * 60))**2) # Eddy       losses in stator yoke
        P_Fesynom = P_Hyys + P_Ftys
        
        P_Hyd     = M_Fest * (B_tmax / 1.5)**2 * (P_Fe0h * om_e / (2 * pi * 60))       # Hysteresis losses in stator teeth
        P_Ftd     = M_Fest * (B_tmax / 1.5)**2 * (P_Fe0e * (om_e / (2 * pi * 60))**2)  # Eddy       losses in stator teeth
        P_Festnom = P_Hyd + P_Ftd
        
        # additional stray losses due to leakage flux
        P_ad = 0.2 * (P_Hyys + P_Ftys + P_Hyd + P_Ftd ) 
        pFtm = 300 # specific magnet loss
        P_Ftm = pFtm * 2 * p * b_m * l_s
        
        Losses = P_Cu + P_Festnom + P_Fesynom + P_ad + P_Ftm
        gen_eff = machine_rating * 100 / (machine_rating + Losses)
        
        #################################################### Structural  Design ############################################################
        
        ## Deflection Calculations ##
        # rotor structure calculations
        
        a_r         = (b_r * d_r) - ((b_r - 2 * t_wr) * (d_r - 2 * t_wr))  # cross-sectional area of rotor arms
        A_r         = L_t * t                                              # cross-sectional area of rotor cylinder
        N_r         = np.round(n_r)                                        # rotor arms
        theta_r     = pi * 1 / N_r                                         # half angle between spokes
        I_r         = L_t * t**3 / 12                                      # second moment of area of rotor cylinder
        I_arm_axi_r = ((b_r * d_r**3) - ((b_r - 2 * t_wr) * (d_r - 2 * t_wr)**3)) / 12  # second moment of area of rotor arm
        I_arm_tor_r = ((d_r * b_r**3) - ((d_r - 2 * t_wr) * (b_r - 2 * t_wr)**3)) / 12  # second moment of area of rotot arm w.r.t torsion
        R           = rad_ag - ag_len - h_m - 0.5 * t                         # Rotor mean radius
        c           = R / 500                                              
        u_all_r     = c / 20                                               # allowable radial deflection
        R_1         = R - t * 0.5                                          # inner radius of rotor cylinder
        k_1         = sqrt(I_r / A_r)                                      # radius of gyration
        m1          = (k_1 / R)**2                                         
        l_ir        = R                                                    # length of rotor arm beam at which rotor cylinder acts
        l_iir       = R_1
        
        
        b_all_r     = 2 * pi * R_o / N_r                                   # allowable circumferential arm dimension for rotor
        q3          = B_g**2 / 2 / mu_0                                    # normal component of Maxwell stress
        mass_PM     = 2 * pi * (R + 0.5 * t) * L_t * h_m * ratio * rho_PM  # magnet mass
        
        # Calculating radial deflection of the rotor
        Numer = R**3 * ((0.25 * (sin(theta_r) - (theta_r * cos(theta_r))) / (sin(theta_r))**2) - (0.5 / sin(theta_r)) + (0.5 / theta_r))
        Pov   = ((theta_r / (sin(theta_r))**2) + 1 / tan(theta_r)) * ((0.25 * R / A_r) + (0.25 * R**3 / I_r))
        Qov   = R**3 / (2 * I_r * theta_r * (m1 + 1))
        Lov   = (R_1 - R_o) / a_r
        Denom = I_r * (Pov - Qov + Lov) # radial deflection % rotor        
        u_Ar  = (q3 * R**2 / E / t) * (1 + Numer / Denom)
        
        # Calculating axial deflection of the rotor under its own weight
        
        w_r         = rho_Fes * g1 * sin(phi) * a_r * N_r                     # uniformly distributed load of the weight of the rotor arm
        mass_st_lam = rho_Fe * 2*pi * R * L_t * h_yr                          # mass of rotor yoke steel
        W           = g1 * sin(phi) * (mass_st_lam / N_r + mass_PM / N_r)     # weight of 1/nth of rotor cylinder
        
        y_a1 = W   * l_ir**3  / 12 / E / I_arm_axi_r                          # deflection from weight component of back iron
        y_a2 = w_r * l_iir**4 / 24 / E / I_arm_axi_r                          # deflection from weight component of the arms
        y_Ar = y_a1 + y_a2 # axial deflection
        
        y_all     = 2 * L_t / 100    # allowable axial deflection
        
        # Calculating # circumferential deflection of the rotor
        
        z_all_r     = radians(0.05  * R)                                   # allowable torsional deflection
        z_A_r       = (2 * pi * (R - 0.5 * t) * L_t / N_r) * sigma \
                      * (l_ir - 0.5 * t)**3 / 3 / E / I_arm_tor_r          # circumferential deflection
        
        val_str_rotor = mass_PM + (mass_st_lam + (N_r * (R_1 - R_o) * a_r * rho_Fes))           # rotor mass
        
        # stator structure deflection calculation
        a_s         = (b_st * d_s) - ((b_st - 2 * t_ws) * (d_s - 2 * t_ws)) # cross-sectional area of stator armms
        A_st        = L_t * t_s                                             # cross-sectional area of stator cylinder
        N_st        = np.round(n_s)                                         # stator arms
        theta_s     = pi * 1 / N_st                                         # half angle between spokes
        I_st        = L_t * t_s**3 / 12                                     # second moment of area of stator cylinder
        k_2         = sqrt(I_st / A_st)                                     # radius of gyration
        I_arm_axi_s = ((b_st * d_s**3) - ((b_st - 2 * t_ws) * (d_s - 2 * t_ws)**3)) / 12  # second moment of area of stator arm
        I_arm_tor_s = ((d_s * b_st**3) - ((d_s - 2 * t_ws) * (b_st - 2 * t_ws)**3)) / 12  # second moment of area of rotot arm w.r.t torsion
        R_st        = rad_ag + h_s + h_ys * 0.5                                # stator cylinder mean radius
        R_1s        = R_st - t_s * 0.5                                      # inner radius of stator cylinder, m
        m2          = (k_2 / R_st)**2
        d_se        = dia + 2 * (h_ys + h_s + h_w)                          # stator outer diameter
        
        # allowable radial deflection of stator
        c1      = R_st / 500
        u_all_s = c1 / 20
        R_out   = (R / 0.995 + h_s + h_ys)
        l_is    = R_st - R_o              # distance at which the weight of the stator cylinder acts
        l_iis   = l_is                    # distance at which the weight of the stator cylinder acts
        l_iiis  = l_is                    # distance at which the weight of the stator cylinder acts
        
        mass_st_lam_s = M_Fest + pi * L_t * rho_Fe * ((R_st + 0.5 * h_ys)**2 - (R_st - 0.5 * h_ys)**2)
        W_is          = 0.5 * g1 * sin(phi) * (rho_Fes * L_t  *d_s**2)                      # length of stator arm beam at which self-weight acts
        W_iis         = g1 * sin(phi) * (mass_st_lam_s + V_Cus * rho_Copper) / 2/N_st       # weight of stator cylinder and teeth
        w_s           = rho_Fes * g1 * sin(phi) * a_s * N_st                                # uniformly distributed load of the arms
                
        mass_stru_steel  = 2 * (N_st * (R_1s - R_o) * a_s * rho_Fes)                        # Structural mass of stator arms
        
        # Calculating radial deflection of the stator
        
        Numers = R_st**3 * ((0.25 * (sin(theta_s) - (theta_s * cos(theta_s))) / (sin(theta_s))**2) - (0.5 / sin(theta_s)) + (0.5 / theta_s))
        Povs   = ((theta_s / (sin(theta_s))**2) + 1 / tan(theta_s)) * ((0.25 * R_st / A_st) + (0.25 * R_st**3 / I_st))
        Qovs   = R_st**3 / (2 * I_st * theta_s * (m2 + 1))
        Lovs   = (R_1s - R_o) * 0.5 / a_s
        Denoms = I_st * (Povs - Qovs + Lovs)
        u_As   = (q3 * R_st**2 / E / t_s) * (1 + Numers / Denoms)

        # Calculating axial deflection of the stator
        X_comp1 = W_is  * l_is**3   / 12 / E / I_arm_axi_s  # deflection component due to stator arm beam at which self-weight acts
        X_comp2 = W_iis * l_iis**4  / 24 / E / I_arm_axi_s  # deflection component due to 1 / nth of stator cylinder
        X_comp3 = w_s   * l_iiis**4 / 24 / E / I_arm_axi_s  # deflection component due to weight of arms
        y_As    = X_comp1 + X_comp2 + X_comp3               # axial deflection
        
        # Calculating circumferential deflection of the stator
        z_A_s   = 2 * pi * (R_st + 0.5 * t_s) * L_t / (2 * N_st) * sigma * (l_is + 0.5 * t_s)**3 / 3/E / I_arm_tor_s 
        z_all_s = radians(0.05 * R_st)               # allowable torsional deflection
        b_all_s = 2 * pi * R_o / N_st                # allowable circumferential arm dimension
        
        val_str_stator = mass_stru_steel + mass_st_lam_s   
        val_str_mass   = val_str_rotor + val_str_stator
        
        TC1 = Torque / (2 * pi * sigma)  # Desired shear stress
        TC2 = R**2 * L_t            # Evaluating Torque constraint for rotor
        TC3 = R_st**2 * L_t         # Evaluating Torque constraint for stator
        
        Structural_mass = mass_stru_steel + (N_r * (R_1 - R_o) * a_r * rho_Fes)
        Stator = mass_st_lam_s + mass_stru_steel + Copper
        Rotor = ((2 * pi * t*L_t * R * rho_Fe) + (N_r * (R_1 - R_o) * a_r * rho_Fes)) + mass_PM
        Mass = Stator + Rotor
        
        
        # Calculating mass moments of inertia and center of mass
        
        I = np.zeros(3)
        I[0]   = 0.50 * Mass * R_out**2
        I[1]   = 0.25 * Mass * R_out**2 + Mass * l_s**2 / 12
        I[2]   = I[1]
        cm = np.zeros(3)
        cm[0]  = shaft_cm[0] + shaft_length / 2. + l_s / 2
        cm[1]  = shaft_cm[1]
        cm[2]  = shaft_cm[2]

        outputs['B_symax']           =  B_symax
        outputs['B_tmax']            =  B_tmax
        outputs['B_rymax']           =  B_rymax
        outputs['B_smax']            =  B_smax
        outputs['B_pm1']             =  B_pm1
        outputs['B_g']               =  B_g
        outputs['N_s']               =  N_s
        outputs['b_s']               =  b_s
        
        outputs['b_t']               =  b_t
        outputs['A_Cuscalc']         =  A_Cuscalc
        outputs['b_m']               =  b_m
        outputs['p']                 =  p
        outputs['E_p']               =  E_p
        outputs['f']                 =  f

        outputs['I_s']               =  I_s
        outputs['R_s']               =  R_s
        outputs['L_s']               =  L_s
        outputs['A_1']               =  A_1
        outputs['J_s']               =  J_s
        outputs['Losses']            =  Losses

        outputs['K_rad']             =  K_rad
        outputs['gen_eff']           =  gen_eff
        outputs['S']                 =  S
        outputs['Slot_aspect_ratio'] =  Slot_aspect_ratio
        outputs['Copper']            =  Copper
        outputs['Iron']              =  Iron
        outputs['u_Ar']              =  u_Ar
        outputs['y_Ar']              =  y_Ar

        outputs['z_A_r']             =  z_A_r
        outputs['u_As']              =  u_As
        outputs['y_As']              =  y_As
        outputs['z_A_s']             =  z_A_s
        outputs['u_all_r']           =  u_all_r
        outputs['u_all_s']           =  u_all_s

        outputs['y_all']             =  y_all
        outputs['z_all_s']           =  z_all_s
        outputs['z_all_r']           =  z_all_r
        outputs['b_all_s']           =  b_all_s
        outputs['b_all_r']           =  b_all_r
        outputs['TC1']               =  TC1

        outputs['TC2']               =  TC2
        outputs['TC3']               =  TC3
        outputs['R_out']             =  R_out
        outputs['Structural_mass']   =  Structural_mass
        outputs['Mass']              =  Mass
        outputs['mass_PM']           =  mass_PM
        outputs['cm']                =  cm
        outputs['I']                 =  I
        

